# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

'''Module for the ``graph`` subcommand.'''

import os
import logging

from ..common import JobCommand, build_graphs


LOGGER = logging.getLogger(__name__)


class GraphCommand(JobCommand):
    '''Command class for the ``build`` subcommand.'''

    NAME = 'graph'

    HELP = 'Output the dependency graph in any format supported by Graphviz.'

    DESC = HELP

    def register(self, parser):
        '''Register options for this command in the parser.'''
        super().register(parser)
        parser.add_argument('--dependencies', action='store',
                            help='select which dependencies to draw (default: '
                            'both)', choices=['hard', 'soft', 'both'],
                            default='both')
        parser.add_argument('-o', '--output',
                            action='store', help='path to the output file')
        parser.set_defaults(func=self.execute)

    @staticmethod
    def execute(args, _config):
        '''Execute the ``graph`` command.'''
        hard_graph, soft_graph = build_graphs(args)
        if args.dependencies == 'hard':
            graph_str = hard_graph.to_graphviz()
        elif args.dependencies == 'soft':
            graph_str = soft_graph.to_graphviz()
        elif args.dependencies == 'both':
            hard_graph_str = hard_graph.to_graphviz()
            soft_graph_str = soft_graph.to_graphviz()
            graph_str = GraphCommand.merge_graph_str(hard_graph_str,
                                                     soft_graph_str)

        if args.output:
            _, ext = os.path.splitext(args.output)
            if ext:
                try:
                    from pydot import graph_from_dot_data
                except ImportError:
                    LOGGER.error('you need to install pydot to use the '
                                 '--output option.')
                    raise
                graph_pydot = graph_from_dot_data(graph_str)[0]
                try:
                    writer = getattr(graph_pydot, 'write_' + ext[1:].lower())
                except AttributeError as aerr:
                    raise ValueError(f'graph output format {ext} is not '
                                     'supported') from aerr
                writer(args.output)
            else:
                with open(args.output, 'w', encoding='utf-8') as fout:
                    fout.write(graph_str)
        else:
            print(graph_str)

    @staticmethod
    def merge_graph_str(solid_graph, dashed_graph):
        r'''Merge two strings representing graphviz graphs into one graph. The
        graphs are merged in such a way that the edges of `solid_graph` and
        `dashed_graph` are respectively represented as solid and dashed.

        .. doctest:: merge_graph_str
            :hide:

            >>> solid_graph = ('digraph {\n  node1 -> node2;\n'
            ... '  node2 -> node3;\n}')
            >>> dashed_graph = 'digraph {\n  node1 -> node3;\n}'

        For example:

        >>> print(solid_graph)
        digraph {
          node1 -> node2;
          node2 -> node3;
        }
        >>> print(dashed_graph)
        digraph {
          node1 -> node3;
        }
        >>> print(GraphCommand.merge_graph_str(solid_graph, dashed_graph))
        digraph {
          subgraph G1 {
          node1 -> node2;
          node2 -> node3;
          }
          subgraph G2 {
          edge [style=dashed];
          node1 -> node3;
          }
        }

        :param str solid_graph: a graph, as produced by
            :meth:`~.DepGraph.to_graphviz`.
        :param str dashed_graph: another graph, as produced by
            :meth:`~.DepGraph.to_graphviz`.
        :returns: the merged graph.
        :rtype: str
        '''
        from itertools import chain
        lines_solid = solid_graph.split('\n')
        lines_dashed = dashed_graph.split('\n')
        lines = list(chain(('digraph {\n  subgraph G1 {',),
                           lines_solid[1:-1],
                           ('  }\n  subgraph G2 {\n  edge [style=dashed];',),
                           lines_dashed[1:-1],
                           ('  }\n}',)))
        return '\n'.join(lines)



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

# pylint: disable=too-many-lines
# pylint: disable=invalid-name
r'''This module provides `pyparsing` grammar for Tripoli-4 output listings.

.. role :: parsing_var(literal)
.. |keff| replace:: k\ :sub:`eff`
.. |kij| replace:: k\ :sub:`ij`
.. _pyparsing: https://pythonhosted.org/pyparsing/

Documentation on the ``pyparsing`` package can be found at `pyparsing`_.

Transformation from ``pyparsing.ParseResults`` to more standard python objects,
including :obj:`numpy` arrays, is done with :mod:`~.transform`, calling
:mod:`~valjean.eponine.tripoli4.common`.

Generalitites
-------------

* This parser only parses the result part of the listing (selection done in
  :mod:`~valjean.eponine.tripoli4.scan`).
* It takes into account all responses in ``qualtrip`` database up to Tripoli-4,
  version 10.2.
* If a response is not taken into account parsing will fail:

  * with a big, ugly message ending by location of the end of successful
    parsing in the result string (possible to print it) → normally where starts
    your new response
  * it seems to end normally, but did not in reality. One of the best checks in
    that case is to test if the ``endflag`` in
    :mod:`~valjean.eponine.tripoli4.scan` was read in the parser, usually not.
    Then the new response probably have to be added.

* A general parser is proposed for use in the file, but other parsers can be
  built from the partial parsers written here
* Numbers are automatically converted to :obj:`numpy` numbers (possibility to
  choose the dtype used for numbers)
* Keywords and most of the variables used to build parsers are private


Organisation
------------

This module is divided in 3 parts:

keywords:
  list of all keywords used to parse the listings, this part is
  important as these keywords trigger the parsing
parsers:
  parsers for each part of the listing (introduction, mesh,
  spectra, general responses, |keff|, etc.)
general parser:
  parser to parse the full listing, taking into accout all
  current response (present in V&V)

Keywords are in most of the cases used as flags and suppressed when data are
stored.

A first structure is designed when building the parsers results as lists and
dictionaries in the ``pyparsing.ParseResults``. Then `parse actions` are used
to standard python or :obj:`numpy` objects. These `parse actions`, called with
``pyparsing.ParserElement.setParseAction``, can be found in :mod:`~.transform`.

Main parsers blocks
```````````````````
The main parsers blocks are defined at the end of the module, named
:parsing_var:`mygram` and :parsing_var:`response`. The default parser is
:parsing_var:`mygram`.

Typically, each result block in the listing should start by the `intro` block,
parsed by :parsing_var:`intro`, and end with at least one `runtime` block,
parsed by :parsing_var:`runtime`. This parts follows the
:mod:`~valjean.eponine.tripoli4.scan`: :obj:`str` starting by
``'RESULTS ARE GIVEN'`` and ending with ``'simulation time'``,
``'exploitation time'`` or ``'elapsed time'``.

Between these blocks can be found the data blocks. The major ones are:

* one or more responses, driven by the keyword ``'RESPONSE FUNCTION'``,
* the editions of IFP adjoint criticality,
* the "default" |keff| block, in most of the cases at the end of the listing,
* the *contributing particles block*, mainly in first pass listings,
* the perturbation block,
* an optional additional `runtime` block.

Main data blocks are described below (results taken into account, main
features).

Response block, parser :parsing_var:`response`
``````````````````````````````````````````````
The core of the listings is the list of responses, including all the required
scores. This big block is constructed as a :obj:`list` of :obj:`dict`, each
one representing a response (key ``'list_responses'`` in the final result).

Response are constructed as:

* response introduction containing its definition, parsed by
  :parsing_var:`respintro`:

  * a description of the response parsed by :parsing_var:`respdesc` including:

    * ``'RESPONSE FUNCTION'`` keyword as mandatory (afaik)
    * ``'RESPONSE NAME'``, ``'SCORE NAME'`` and ``'ENERGY DECOUPAGE NAME'``
      that are present in most of the cases

  * more characteristics of the response, parsed by :parsing_var:`respcarac`,
    like:

    * considered particle (``'PARTICULE'`` in the listing)
    * nucleus on which the reaction happens (if ``'RESPONSE FUNCTION'`` is a
      ``'REACTION'``)
    * temperature
    * compostion of the volumes considered
    * concentration
    * reaction considered (usually given as codes)
    * others like DPA type, required arguments, mode, filters, etc.

* responses themselves, using parser :parsing_var:`responseblock`, are various:

  * responses including *score* description, all included in the
    :parsing_var:`scoreblock` parser. More than one can be present, they are
    grouped in the :parsing_var:`listscoreblock` parser.
    :parsing_var:`scoreblock` parser contains:

    * score description (parser :parsing_var:`scoredesc`) contains the score
      mode (``'TRACK'``, ``'SURF'`` or ``'COLL'``) and the score zone
      (currently taken into account: mesh, results cumulated on all geometry or
      on all sources, Volume, Volume Sum, Frontier, Frontier Sum, Point, Cells
      and Maille)

    * results block, where at least one of these results can be found, are by
      parsed by the following parsers:

      * :parsing_var:`spectrumblock`: spectrum
      * :parsing_var:`meshblock`: mesh
      * :parsing_var:`vovspectrumblock`: spectrum with variance of variance
      * :parsing_var:`entropy`: entropy results (Boltzmann and Shannon
        entropies)
      * :parsing_var:`medfile`: location of optional MED file
      * :parsing_var:`genericscoreblock`: default result integrated over energy
      * :parsing_var:`uncertblock`: uncertainty results
      * :parsing_var:`uncertintegblock`: uncertainties on integrated results
        over energy
      * :parsing_var:`gbblock`: Green bands results

  * |keff| presented as a generic response, possibly transformed in
    :obj:`numpy.matrix` (parser :parsing_var:`keffblock`)
  * |kij| results: matrix, eigenvalues, eigenvectors (parser
    :parsing_var:`kijres`)
  * |kij| sources (parser :parsing_var:`kijsources`)
  * Adjoint related results (parser :parsing_var:`adjointres`): scores ordered
    by precursors and families, by perturbation index, by cycle length or
    sensitivities (this last case is represented in a 3 dimensions
    :obj:`numpy.ndarray`, incident energy, energy ("leaving neutron"),
    direction cosine (µ)). For the moment this is only for IFP method , in
    close future also for Wielandt method
  * default result integrated over energy where no scoring mode and zone are
    precised (parser :parsing_var:`defintegratedres`)
  * perturbation results (parser :parsing_var:`perturbation`)


Other parsers
`````````````
Various other blocks can appear in the Tripoli-4 listing, located at the same
level as the response block. These parsers and the associated dictionary key
(same level as ``'list_responses'``) are:

* :parsing_var:`ifpadjointcriticality`: edition of IFP adjoint criticality, key
  ``'ifp_adjoint_crit_edition'``;
* :parsing_var:`autokeffblock`: "default" |keff| block, containing for example
  the best estimation of |keff| using variable number of discarded batches, key
  ``'keff_auto'``;
* :parsing_var:`contribpartblock`: *contributing particles block*, key
  ``'contributing_particles'``
* :parsing_var:`perturbation`: perturbation results, containing a description
  of the calculation of the perturbation followed by the perturbation result
  presented like a usual response (spectrum, mesh, etc. depending on required
  score), key ``'perturbation'``
* :parsing_var:`runtime`: simulation, exploitation or elapsed time.


.. todo::

    Adjoint results: for the moment only IFP is really parsed. Grammar has
    already more or less adapted to welcome Wielandt method that will have the
    same kind of outputs (renaming as adjoint_res for example). No key is set
    for the moment to specify the method, it can be obtained from the response
    function itself. Adjoint criticality editions are only done for IFP, this
    may change when the same will be available for Wielandt. Some renaming can
    also be needed.

'''

import logging
from pyparsing import (Word, Keyword, White, alphas, alphanums, LineEnd,
                       Suppress, Optional, CaselessKeyword, original_text_for,
                       Group, OneOrMore, ZeroOrMore, Forward,
                       token_map, delimited_list, printables, replace_with)
from pyparsing import pyparsing_common as pyparscom
from . import transform as trans
from .transform import compose2
from .dump import dump_in_logger


LOGGER = logging.getLogger(__name__)

_fnums = pyparscom.fnumber.set_parse_action(token_map(trans.common.FTYPE))
_inums = pyparscom.number.set_parse_action(token_map(trans.common.ITYPE))

###################################
#            KEYWORDS             #
###################################

# General keywords
_integratedres_kw = Keyword("INTEGRATED RESULTS")
_numbatchsused_kw = (Keyword("number of")
                     + (Keyword("batches") | Keyword("batch"))
                     + Optional(Keyword("used")))
_numbatchs1stdiscarded_kw = Keyword("number of first discarded batches")
_notconverged_kw = (Keyword("NOT YET CONVERGED")
                    | CaselessKeyword("Not converged"))
_unknown_kw = Keyword("unknown")
_unavailable_kw = Keyword("unavailable")
_units_kw = Keyword("Units:")
_warning_kw = CaselessKeyword("Warning")
_endtable = LineEnd() + LineEnd()

# Introduction keywords
_sourceintensity_kw = Keyword("RESULTS ARE GIVEN FOR SOURCE INTENSITY")
_meanweightleakage_kw = Keyword("Mean weight leakage")
_meanweightleakagein_kw = Keyword("Mean weight leakage inside")
_edbatchnum_kw = Keyword("Edition after batch number")
_meanweightrestartpart_kw = Keyword("Mean weight of restarted particles :")

# End of edition keywords
_simulationtime_kw = Keyword("simulation time (s)")
_exploitationtime_kw = Keyword("exploitation time (s)")
_elapsedtime_kw = Keyword("elapsed time (s)")
_rdmgenerator_kw = Keyword("Type and parameters of random generator "
                           "at the end of simulation")
# _rdmgenerator_kw = Keyword("Type and parameters of random generator" +
# ("at the end of simulation"|"after batch"))
_normalcompletion_kw = Keyword("NORMAL COMPLETION")

# Response description keywords
_respfunction_kw = Keyword("RESPONSE FUNCTION")
_respname_kw = Keyword("RESPONSE NAME")
_scorename_kw = Keyword("SCORE NAME")
_energysplitname_kw = Keyword("ENERGY DECOUPAGE NAME")
_nusplitname_kw = Keyword("DECOUPAGE NAME")
_respfiltered_kw = Keyword("RESPONSE FILTERED BY")
_respfiltcompos_kw = Keyword("COMPOSITIONS")
_particule_kw = Keyword("PARTICULE")
_incparticle_kw = Keyword("INCIDENT PARTICULE")
_noise_equation_kw = Keyword("NOISE EQUATION")
_reactiononnucl_kw = Keyword("reaction on nucleus")
_temperature_kw = Keyword("temperature")
_composition_kw = Keyword("composition")
_concentration_kw = Keyword("concentration")
_reaction_kw = Keyword("reaction consists in")
_required_kw = Keyword("REQUIRED")
_dpatype_kw = Keyword("DPA TYPE:")
_mode_kw = Keyword("MODE :")
_inducedbyint_kw = Keyword("INDUCED BY INTERACTION :")
_fxptcontrib_kw = Keyword("FXPT CONTRIBUTION")
_spectrumresp_kw = Keyword("SPECTRUM")
_filters_kw = Keyword("Score filtered by volume")

# Scoring description
_scoremode_kw = Keyword("scoring mode")
_scorezone_kw = Keyword("scoring zone")
_scoremesh_kw = (Keyword("Results on a mesh")
                 .set_parse_action(replace_with('Mesh')))
_scoremeshinfobins_kw = Keyword("(x,y,z)")
_scoreallgeom_kw = (Keyword("Results cumulated on all geometry")
                    .set_parse_action(replace_with('All geometry')))
_scoreallsources_kw = (Keyword("Results cumulated on all sources")
                       .set_parse_action(replace_with('All sources')))
_scorevol_kw = Keyword("Volume")
_scorevolvol_kw = Keyword("num of volume")
_scorevolumeunit_kw = Keyword("Volume in")
_scorevolumeint_kw = (Keyword("The result is integrated over the volume")
                      | Keyword("The result is integrated in volume")
                      ).set_parse_action(replace_with('volume integrated'))
_scorevolumenotint_kw = (
    Keyword("The volume has been provided by the user "
            "(the user requested a score per unit volume)")
    | Keyword("The volume has been calculated by Tripoli-4 "
              "or provided by the user")).set_parse_action(
                  replace_with('per unit volume'))
_scorevolsum_kw = Keyword("Volume Sum")
_scorevolsumvol_kw = Keyword("num of volumes")
_scorevolumesum_kw = Keyword("Total volume in cm3")
_scorevolumecell_kw = Keyword("Cell volume in cm3")
_scoresurf_kw = Keyword("Frontier")
_scoresurfvol_kw = Keyword("volumes")
_scoresurface_kw = Keyword("Surface in cm2")
_scoresurfaceint_kw = Keyword("The surface area has been provided by the user "
                              "(the user requested a score per unit area)")
_scoresurfacenotint_kw = Keyword("The result is integrated over the surface")
_scoresurfsum_kw = Keyword("Frontier Sum")
_scoresurfsumfront_kw = Keyword("num of frontiers")
_scoresurfacesum_kw = Keyword("Total surface in cm2")
_scorepoint_kw = Keyword("Point")
_scorecell_kw = Keyword("Cells")
_scorecelldet_kw = Keyword("(numvol,depth,imaille,jmaille,kmaille...)")
_scoremaille_kw = Keyword("Maille")
_scoremaillevol_kw = Keyword("num of volume")
_scoremailledepth_kw = Keyword("depth of lattice")
_scoremaillecell_kw = Keyword("num of cell")

# Correspondence table (volumes and their names)
_corresptable_kw = Keyword("Correspondence table between volumes "
                           "ids and names :")
_vol_kw = CaselessKeyword("Volume")

# KEFF keywords
_notcvg_exp_kw = Keyword("(invalid keff domain)")
_fullcomb_kw = Keyword("full combined estimator")
_bestresdiscbatchs_kw = Keyword("best results are obtained with discarding")
_correlations_kw = Group(Keyword("estimators")
                         + Keyword("correlations")
                         + Keyword("combined values")
                         + Keyword("combined sigma%"))
_estimator_kw = Keyword("ESTIMATOR")
_equivkeff_kw = Keyword("Equivalent Keff:")
_warn_combkeff_kw = (
    Keyword("One of the Keffectives is null and should not be")
    + Keyword("Combined Keffectives will not be edited"))
_warn_fixsourcekeff_kw = (
    Keyword("In FIXED_SOURCES_CRITICITY mode, the keff result")
    + Keyword("is actually an overall multiplication factor "
              "(cf User's Guide)"))
_parna_likelihood_kw = Keyword("parna likelihood confidence interval (on the "
                               "mean of M = M' + 1)")
_parna_inv_cdf_kw = Keyword("inversion of the CDF converged after")

# Time steps
_timestepnum_kw = Keyword("TIME STEP NUMBER")
_timestepmin_kw = Keyword("time min. =")
_timestepmax_kw = Keyword("time max. =")

# Angular zones
_muangzone_kw = Keyword("MU ANGULAR ZONE :")
_mumin_kw = Keyword("mu min. =")
_mumax_kw = Keyword("mu max. =")
_phiangzone_kw = Keyword("PHI ANGULAR ZONE :")
_phimin_kw = Keyword("phi min. =")
_phimax_kw = Keyword("phi max. =")

# Spectrum keywords
_spectrum_kw = Keyword("SPECTRUM RESULTS")
_spgroupwunit_kw = Keyword("group (MeV)")
_spgroup_kw = Keyword("group")
_spscovlethargy_kw = Keyword("score/lethargy")
_spvov_kw = Keyword("vov")
_nuspectrum_kw = Keyword("NU RESULTS")
_nusprange_kw = Keyword("range")
_zaspectrum_kw = Keyword("ZA RESULTS")
_zaspid_kw = Keyword("(Z,A)")
_spscore_kw = Keyword("score")
_spsigma_kw = Keyword("sigma_%")

# Mesh keywords
_energyrange_kw = Keyword("Energy range")

# MED files
_creationmedfile_kw = Keyword("Creating MED output file")
_creationfile_kw = Keyword("# Creating output file")
_medmeshid_kw = Keyword("MED mesh id")

# Entropy
_boltzmannentropy_kw = Keyword("Boltzmann Entropy of sources =")
_shannonentropy_kw = Keyword("Shannon Entropy of sources =")

# Scores ordered by nuclei and precursor families
_nucleiorder_kw = Keyword("Scores for nuclei contributions are ordered "
                          "according to the user list:")
_familiesorder_kw = Keyword("Scores are ordered from family i = 1 to i = MAX:")
_nucleifamilyorder_kw = Keyword("Scores are ordered "
                                "by nuclei and by families:")
_nucleus_kw = Keyword("Nucleus :")
# Perturbation order
_perturborder_kw = Keyword("Scores are ordered by perturbation index:")
# Sensitivities
_sensitivitytypeorder_kw = Keyword("Scores are ordered by type")
_sensitivityindexorder_kw = Keyword("and index:")
_sensitivity_kw = Keyword("SENSITIVITY :")
_sensitivityreactionrate_kw = Keyword("REACTION_RATE_RATIO :")
_sensitivity_energyint_kw = Keyword("Energy integrated S")
_sensitivity_incenergy_kw = Keyword("Incident energy interval in MeV:")
_sensitivity_dircos_kw = Keyword("Direction cosine interval:")

# Variance of variance
_vovstar_kw = Keyword("variance of variance* :")
_sensibtomaxval_kw = Keyword("sensibility to maximum value:")
_vov_kw = Keyword("variance of variance :")

# Greenbands
_gbspectrumstep_kw = Keyword("* SOURCE SPECTRUM STEP NUMBER :")
_gbenergymin_kw = Keyword("source energy min. =")
_gbenergymax_kw = Keyword("source energy max. =")
_gbsourcenum_kw = Keyword("SOURCE NUMBER :")
_gbsourcetab_kw = Keyword("SOURCE TABULATION :")

# KIJ matrix
_kijlefteigenval_kw = Keyword("left_eigenvalues called")
_kijmkeff_kw = Keyword("kij-keff =")
_kijdomratio_kw = Keyword("dominant ratio =")
_kijeigenval_kw = Keyword("eigenvalues (re, im)")
_kijeigenvec_kw = Keyword("eigenvectors")
_kijeigenvecnotprint_kw = Keyword("KIJ eigenvectors not printed, "
                                  "increase maximum dump size if needed")
_kijmatrix_kw = Keyword("KIJ_MATRIX :")
_kijmatrixnotprint_kw = Keyword("KIJ matrix not printed, "
                                "increase maximum dump size if needed")

# KIJ sources
_kijsources_kw = Keyword("SOURCES VECTOR :")
_kijsourcesorder_kw = Keyword("Sources are ordered following")

# KIJ keff
_kijfissilevol_kw = Keyword("number of fissile volumes :")
_kijlistfissilevol_kw = Keyword("list of fissile volume numbers : ")
_kijbatchs_kw = Keyword("number of last batches kept :")
_kijkeffevid_kw = (Keyword("EIGENVECTOR :") + Keyword("index")
                   + Keyword("source rate"))
_kijkeffmat_kw = Keyword("K-IJ MATRIX :")
_kijkeffstddevmat_kw = Keyword("STANDARD DEVIATION MATRIX :")
_kijkeffsensibilitymat_kw = Keyword("SENSIBILITY MATRIX :")

# Adjoint results (IFP for the moment)
# if IFP is changed to Wielandt in _cvgstat_kw when using Wielandt method this
# is the way to get the method name
# convergence statistics
_cvgstat_kw = Keyword("Scores for IFP convergence statistics are ordered "
                      "from cycle length L = 1 to L = MAX:")
# IFP adjoint criticality edition
_ifpadjcriticality_kw = Keyword("IFP_ADJOINT_CRITICALITY EDITION")
_ifpadjcyclelength_kw = Keyword("IFP CYCLE LENGTH =")
_ifpadjnormalizedres_kw = Keyword("RESULTS ARE NORMALIZED")
_ifpadjvol_kw = Keyword("Vol")
_ifpadjminmax_kw = Keyword("(min | max)")
_ifpadjscore_kw = Keyword("score [a.u.]")

# Results on spherical harmonics
_shr_nb_space_bins_kw = Keyword('Number of space bin :')
_shr_nb_u_space_bins_kw = Keyword('Number of space bin in u dimension:')
_shr_nb_v_space_bins_kw = Keyword('Number of space bin in v dimension:')
_shr_nb_w_space_bins_kw = Keyword('Number of space bin in w dimension:')
_shr_nb_energy_bins_kw = Keyword('Number of energetic bins :')
_shr_nb_inc_energy_bins_kw = Keyword('Number of incident energy bins :')
_shr_lmax_kw = Keyword('Maximum L number of moments :')
_shr_score_name_kw = Keyword('Score :')
_shr_space_kw = Keyword('Space bin :')
_shr_energy_kw = Keyword('Energy range (in MeV):')
_shr_incident_energy_kw = Keyword('Outgoing data for incident energy '
                                  'bin boundaries :')
_shr_l_kw = Keyword('l')
_shr_m_kw = Keyword('m')
_shr_score_kw = Keyword('Score')
_shr_sigma_kw = Keyword('Sigma (%)')

# Perturbations
_perturbation_kw = Keyword("================== Perturbation result edition "
                           "======================")
_perturank_kw = Keyword("Perturbation rank =")
_pertumethod_kw = Keyword("Method :")
_pertuorder_kw = Keyword("Order:")
_pertutype_kw = Keyword("Perturbation de type")
_pertucompo_kw = Keyword("Composition :")

# Uncertainties results (linked to perturbations ?)
_uncertres_kw = Keyword("UNCERTAINTY RESULTS")
_uncertgp_kw = Keyword("group (Mev)")
_uncertsig2_kw = Keyword("sigma2(means)")
_uncertmean_kw = Keyword("mean(sigma_n2)")
_uncertsig_kw = Keyword("sigma(sigma_n2)")
_uncertfisher_kw = Keyword("fisher test")
_uncertintegres_kw = Keyword("UNCERTAINTY ON ENERGY INTEGRATED RESULTS")

# Creation jdds
_nbcontribpart_kw = Keyword("NUMBER OF CONTRIBUTING PARTICLES")
_endcontribpart_kw = Keyword("--- end of CONTRIBUTING PARTICLES ---")

# Symbol lines
_star_line = Suppress(Word('*'))
_equal_line = Suppress(Word('='))
_minus_line = Suppress(White() + Word('-'))


################################
#           PARSERS            #
################################

# Introduction parser
_sourceintensity = (Suppress(_sourceintensity_kw + ':')
                    + (_fnums | _unavailable_kw)
                    ('source_intensity'))
# unkown -> string not in a list, while list of float per default
_meanweightleakvals = Group(_fnums('score')
                            + Suppress("sigma =") + _fnums('sigma')
                            + Suppress('sigma% =') + _fnums('sigma%'))
_meanweightleak = (Suppress(_meanweightleakage_kw)
                   + (Suppress('=') + _meanweightleakvals('mean_weight_leak')
                      | Suppress(':') + _unknown_kw('mean_weight_leak')))
_meanweightleakin = (
    Suppress(_meanweightleakagein_kw)
    + (Suppress('=') + _meanweightleakvals('mean_weight_leak_inside')
       | Suppress(':') + _unknown_kw('mean_weight_leak_inside')))
_edbatchnum = Suppress(_edbatchnum_kw + ':') + _inums('edition_batch_number')

_meanweightrestartpart = (Suppress(_meanweightrestartpart_kw)
                          + _fnums('mean_weight_restart_particle'))
_introelts = (_meanweightleak | _meanweightleakin | _edbatchnum
              | _meanweightrestartpart)
intro = Group(_sourceintensity + _star_line + OneOrMore(_introelts))('intro')

# Conclusion parser
_packet_length_warning = Suppress(
    "* packet length is" + _inums
    + "(check documentation for conventions about discard and batches)")
_simutime = Suppress(_simulationtime_kw + ':') + _fnums('simulation_time')
_exploitime = (Suppress(_exploitationtime_kw + ':')
               + _fnums('exploitation_time'))
_elapsedtime = Suppress(_elapsedtime_kw + ':') + _fnums('elapsed_time')
runtime = (Optional(_packet_length_warning)
           + Group(_simutime | _elapsedtime | _exploitime)('conclu'))

# Response parser
# Description of the response
_respfunc = (Suppress(_respfunction_kw + ':')
             + OneOrMore(Word(printables), stopOn=LineEnd())
             .set_parse_action(' '.join)('response_function'))
# warning: order matters hier, LineEnd has to be before Optional(Word)
_respname = (Suppress(_respname_kw + ':')
             + (Suppress(LineEnd())
                | Optional(Word(printables)('response_name'))))
_scorename = (Suppress(_scorename_kw + ":")
              + Optional(Word(printables)('score_name')))
_energysplit = (Suppress(_energysplitname_kw + ':')
                + Word(printables)('energy_split_name'))
_nusplit = (Suppress(_nusplitname_kw + ':')
            + Word(printables)('nu_split_name'))
respdesc = (_respfunc + Optional(_respname) + Optional(_scorename)
            + Optional(_energysplit | _nusplit))

_respfilter = (Suppress(_respfiltered_kw) + _inums('nb_filtered_compos')
               + Suppress(_respfiltcompos_kw + ':')
               + Group(OneOrMore(Word(printables), stopOn=LineEnd()))
               .set_parse_action(trans.convert_list_to_tuple)
               ('filtered_compositions'))

_particle = (Suppress(_particule_kw + ':')
             + OneOrMore(Word(alphas+','), stopOn=LineEnd())
             .set_parse_action(' '.join)('particle'))
_incparticle = (Suppress(_incparticle_kw + ':')
                + Word(alphas)('incident_particle'))
_noise_equation = (Suppress(_noise_equation_kw + ':')
                   + OneOrMore(Word(alphas), stopOn=LineEnd())
                   .set_parse_action(' '.join)('noise_equation'))
# response characteristics written in lower case
_reactiononnucl = (Suppress(_reactiononnucl_kw + ':')
                   + Word(printables)('reaction_on_nucleus'))
_temperature = Suppress(_temperature_kw + ':') + _fnums('temperature')
_composition = (Suppress(_composition_kw + ':')
                + Word(printables)('composition'))
_concentration = Suppress(_concentration_kw + ':') + _fnums('concentration')
_reaction = (Suppress(_reaction_kw)
             + ((Suppress(Word(alphas) + ':')
                 + Group(delimited_list(_inums, delim='+')))
                | OneOrMore(Word(alphas), stopOn=LineEnd())
                .set_parse_action(' '.join))
             .set_parse_action(trans.convert_list_to_tuple)('reaction'))


# Goal: when more than one reaction are required, keep characteristics grouped
# by particle, reaction, etc.
def _next_compos(toks):
    if toks.getName() == 'reaction_on_nucleus':
        detail = _temperature | _composition | _concentration | _reaction
    elif toks.getName() == 'temperature':
        detail = _reactiononnucl | _composition | _concentration | _reaction
    elif toks.getName() == 'composition':
        detail = _reactiononnucl | _temperature | _concentration | _reaction
    elif toks.getName() == 'concentration':
        detail = _reactiononnucl | _temperature | _composition | _reaction
    elif toks.getName() == 'reaction':
        detail = _reactiononnucl | _temperature | _composition | _concentration
    else:
        LOGGER.warning("Not a foreseen result name, please check, keeping all")
        detail = (_reactiononnucl | _temperature | _composition
                  | _concentration | _reaction)
    _otherdetails << OneOrMore(detail)  # pylint: disable=W0106


_compodetails = Forward()
_otherdetails = Forward()
_compoptions = (_reactiononnucl
                | _temperature
                | _composition
                | _concentration
                | _reaction).set_parse_action(_next_compos)
_compodetails << Group(_compoptions + _otherdetails)  # pylint: disable=W0106
_nuclflags = (OneOrMore(_compodetails).set_parse_action(trans.lod_to_dot)
              ('compos_details'))


# other response characteristics
_dpatype = (Suppress(_dpatype_kw)
            + OneOrMore(Word(alphas+'-,'), stopOn=LineEnd())
            .set_parse_action(' '.join)('dpa_type'))
_required = (Suppress(_required_kw)
             + OneOrMore(Word(alphanums+'():'), stopOn=LineEnd())
             .set_parse_action(' '.join)('required'))
_mode = Suppress(_mode_kw) + Word(alphas)('mode')
_inducedbyint = (Suppress(_inducedbyint_kw)
                 + Group(OneOrMore(_inums))
                 .set_parse_action(trans.convert_list_to_tuple)
                 ('induced_by_interation'))
_notinducedbyint = (Suppress("NOT" + _inducedbyint_kw)
                    + Group(OneOrMore(_inums))
                    .set_parse_action(trans.convert_list_to_tuple)
                    ('NOT_induced_by_interation'))
_fxptcontrib = (OneOrMore(Word(alphas+'()'),
                          stopOn=_fxptcontrib_kw).set_parse_action(' '.join)
                ('fxpt_contribution')
                + Suppress(_fxptcontrib_kw))
_spectrumresp = Suppress(_spectrumresp_kw + ':') + Word(alphas)('spectrum')
_filters = (Suppress(_filters_kw + ':')
            + Group(OneOrMore(_inums))
            .set_parse_action(trans.convert_list_to_tuple)
            ('filtered_volumes')
            + Optional(Suppress('With') + Word(alphas+'_')('filter_type')))


respcarac = (_respfilter
             | _particle
             | _incparticle
             | _noise_equation
             | _nuclflags
             | _filters
             | _required
             | _dpatype
             | _mode
             | _inducedbyint
             | _notinducedbyint
             | _fxptcontrib
             | _spectrumresp)

respintro = respdesc + ZeroOrMore(respcarac)

# Responses themselves
# Score description (not needed for KEFF)
scoremode = Suppress(_scoremode_kw + ':') + Word(alphas+'_')('scoring_mode')
# scoring zones
_score_mesh_unit = (Suppress("(in")
                    + Word(alphanums+'.^-+')('unit')
                    + Suppress(')'))
_score_meshinfo_cell_unit = (Suppress(_scoremeshinfobins_kw + '[')
                             + _fnums + Word(alphanums+' .^-+')
                             + Suppress(']'))('cell_unit')
_score_meshinfo_volume = (Suppress(_vol_kw + '[')
                          + _fnums + Word(alphanums+'.^-+')
                          + Suppress(']'))('vol_unit')
_score_meshcols = Suppress(Word(alphas)
                           + Optional(_scoremeshinfobins_kw)
                           + Word(alphas)
                           + Word(alphas+'() '))
_score_meshinfounit_cols = (Suppress(Word(alphas))
                            + _score_meshinfo_cell_unit
                            + _score_meshinfo_volume
                            + Suppress(Word(alphas) + Word(alphas+'() ')))
_score_mesh = (Optional(Suppress(_scorevolumeunit_kw)
                        + Word(alphanums+' .^-+')('scoring_zone_volsurf_unit')
                        + Suppress(':')
                        + _fnums('scoring_zone_volsurf'))
               + Optional(Suppress(_scorevolumecell_kw + ':')
                          + _fnums('scoring_zone_cellvol'))
               + _scoremesh_kw('scoring_zone_type') + Suppress(':')
               + (_score_meshinfounit_cols | _score_meshcols)
               + Optional(_score_mesh_unit))
_score_allgeom = _scoreallgeom_kw('scoring_zone_type')
_score_allsources = _scoreallsources_kw('scoring_zone_type')
_score_vol = (_scorevol_kw('scoring_zone_type')
              + Suppress(_scorevolvol_kw + ':')
              + (_inums | Word(printables))('scoring_zone_id')
              + Suppress(_scorevolumeunit_kw)
              + Word(alphanums+' .^-+')('scoring_zone_volsurf_unit')
              + Suppress(':')
              + _fnums('scoring_zone_volsurf')
              + Optional(_scorevolumeint_kw | _scorevolumenotint_kw)
              ('scoring_zone_details'))
_score_vol_sum = (
    (_scorevolsum_kw('scoring_zone_type')
     + Suppress(_scorevolsumvol_kw + ':')
     + Group(delimited_list(_inums, delim='+'))('scoring_zone_id')
     + Suppress(_scorevolumesum_kw + ':')
     + _fnums('scoring_zone_volsurf')))
_score_surf = (_scoresurf_kw('scoring_zone_type')
               + Suppress(_scoresurfvol_kw + ':')
               + Group(delimited_list(_inums
                                      | Word(printables, excludeChars=',')))
               ('scoring_zone_id')
               + Optional(Suppress(_scoresurface_kw + ':')
                          + _fnums('scoring_zone_volsurf'))
               + Optional(Suppress(_scoresurfaceint_kw
                                   | _scoresurfacenotint_kw)))
_score_surf_sum = (
    _scoresurfsum_kw('scoring_zone_type')
    + Suppress(_scoresurfsumfront_kw + ':')
    + delimited_list(
        Group(Suppress('(')
              + delimited_list(Word(printables, excludeChars=',()') | _inums)
              + Suppress(')')),
        delim='+')('scoring_zone_id')
    + Suppress(_scoresurfacesum_kw + ':')
    + _fnums('scoring_zone_volsurf'))
_score_point = (_scorepoint_kw('scoring_zone_type')
                + Suppress(':')
                + Group(_fnums + Suppress(',')
                        + _fnums + Suppress(',')
                        + _fnums)('scoring_zone_id'))
_cellelt = (Group(Suppress('(') + _inums
                  + Suppress(',') + _inums
                  + OneOrMore(
                      Group(Suppress(',') + _inums
                            + Suppress(',') + _inums
                            + Suppress(',') + _inums))
                  + Suppress(')')))
_score_cell = (_scorecell_kw('scoring_zone_type')
               + Suppress(_scorecelldet_kw)
               + Group(_cellelt
                       + ZeroOrMore(Suppress('+') + _cellelt))
               ('scoring_zone_id'))
_maillelt = (Suppress(_scoremaillevol_kw + ':') + _inums
             + Suppress(_scoremailledepth_kw + ':') + _inums
             + Suppress(_scoremaillecell_kw + ':')
             + OneOrMore(
                 Group(Suppress('(') + _inums
                       + Suppress(',') + _inums
                       + Suppress(',') + _inums + Suppress(')'))))
_score_maille = (_scoremaille_kw('scoring_zone_type')
                 + Group(_maillelt)('scoring_zone_id'))
scorezone = (Suppress(_scorezone_kw+':')
             + (_score_mesh
                | _score_allgeom
                | _score_allsources
                | _score_vol
                | _score_surf
                | _score_surf_sum
                | _score_vol_sum
                | _score_point
                | _score_cell
                | _score_maille
                | LineEnd()))
# scoring description = scoring mode + scoring zone
scoredesc = scoremode + scorezone

# Correspondence table (volumes ids and names)
corresptable = (Suppress(_corresptable_kw)
                + OneOrMore(Group(
                    Suppress(_vol_kw + ':')
                    + _inums('volume_id')
                    + Suppress(Keyword('is :'))
                    + Word(printables)('volume_name')), stopOn=_endtable)
                ('correspondence_table'))


# RESPONSE RESULTS


def _set_no_unit_case(toks):
    '''Deal with the "not unit" case'''
    if len(toks) == 1:
        return {'uscore': '', 'usigma': toks[0]}
    LOGGER.warning("more than one unit, please check: %s", toks)
    return None


def _rm_blanks(toks):
    '''Remove leading and trailing spaces (not the ones inside the string)'''
    return toks[0].strip()


# Default integrated result
_numdiscbatch = (Suppress(_numbatchs1stdiscarded_kw + ':')
                 + _inums('discarded_batches'))
_numusedbatch = Suppress(_numbatchsused_kw + ':') + _inums('used_batches')
_integratedres_name = _integratedres_kw | Word(alphas) + _integratedres_kw
_integratedres = _fnums('score') + _fnums('sigma')
_unitsres = (
    Suppress(_units_kw)
    + (Word('%').set_parse_action(_set_no_unit_case)
       | (Word(alphanums+'.^-+ ').set_parse_action(_rm_blanks)('uscore')
          + Word('%')('usigma'))))
# rejection in vov and sensibility cases
_rejection = (Suppress('[')
              + OneOrMore(Word(alphanums+'<>.+-')).set_parse_action(' '.join)
              ('rejection')
              + Suppress(']'))
_vovnostar = Suppress(_vov_kw) + Group(_fnums('score'))('vov_res')
_vovstar = Group(
    Suppress(_vovstar_kw) + _fnums('score')
    + Suppress('[') + _fnums('sigma') + Suppress(']')
    + Optional(_rejection))('vovstar_res')
_sensibtomaxval = Group(
    Suppress(_sensibtomaxval_kw) + _fnums('max_val')
    + Optional(_rejection))('sensibility_res')
_vov = _vovnostar | _vovstar + _sensibtomaxval
# best result
bestres = (Group(Suppress(_bestresdiscbatchs_kw) + _inums('discarded_batches')
                 + Suppress("batches")
                 + _minus_line
                 + _numusedbatch + _fnums('score') + _fnums('sigma'))
           ('best_result_res'))


integratedres = (Group(Optional(Suppress(_integratedres_name))
                       + Optional(_numdiscbatch)
                       + ((_numusedbatch + _integratedres)
                          | _notconverged_kw('not_converged')))
                 ('integrated_res')
                 + Optional(_vov))

genericscoreblock = (Group(Optional(Suppress(_integratedres_name))
                           + ((_numusedbatch
                               + _integratedres
                               + Optional(_unitsres))
                              | _notconverged_kw('not_converged')))
                     ('generic_res').set_parse_action(trans.group_to_dict))


# Time steps
_timestep = Group(Suppress(_timestepnum_kw + ':') + _inums
                  + _minus_line
                  + Suppress(_timestepmin_kw) + _fnums
                  + Suppress(_timestepmax_kw) + _fnums)('time_step')


# Angular zones
_muangzone = Group(Suppress(_muangzone_kw) + _inums
                   + _minus_line
                   + Suppress(_mumin_kw) + _fnums
                   + Suppress(_mumax_kw) + _fnums)('mu_angle_zone')
_phiangzone = Group(Suppress(_phiangzone_kw) + _inums
                    + _minus_line
                    + Suppress(_phimin_kw) + _fnums
                    + Suppress(_phimax_kw) + _fnums)('phi_angle_zone')


# Spectrum
_spectrumunits = Group(Suppress(_units_kw)
                       + Word(alphanums+'.^-+%') * 4)('units')
_spectrumbin = _fnums + Suppress('-') + _fnums
_spectrumcols = Suppress((_spgroupwunit_kw | _spgroup_kw)
                         + _spscore_kw
                         + _spsigma_kw
                         + _spscovlethargy_kw)
_spectrumvals = (Group(_spectrumbin + _fnums + _fnums + _fnums)
                 .set_fail_action(trans.fail_spectrum))
_spectrum = (Suppress(_spectrum_kw)
             + _numdiscbatch
             + _spectrumcols
             + Optional(_spectrumunits)
             + OneOrMore(_spectrumvals, stopOn=_endtable)('spectrum_vals'))
spectrumblock = (Group(OneOrMore
                       (Group(OneOrMore(_timestep | _muangzone | _phiangzone)
                              + _spectrum
                              + Optional(integratedres)))
                       | Group(_spectrum))('spectrum_res'))


# Spectrum with vov
# no unit for vov (not even a space) -> use spectrum ones with 4 fields
_vovspectrumcols = Suppress((_spgroupwunit_kw | _spgroup_kw)
                            + _spscore_kw + _spsigma_kw
                            + _spscovlethargy_kw + _spvov_kw)
_vovspectrumbin = _fnums + Suppress('-') + _fnums
_vovspectrumvals = Group(_vovspectrumbin + _fnums + _fnums + _fnums + _fnums)
_vovspectrum = (Suppress(_spectrum_kw)
                + _numdiscbatch
                + _vovspectrumcols
                + Optional(_spectrumunits)
                + OneOrMore(_vovspectrumvals, stopOn=_endtable)
                ('spectrum_vals'))
vovspectrumblock = Group(Group(_vovspectrum))('vov_spectrum_res')


# Nu spectrum
_nuspectrumunits = Group(Suppress(_units_kw)
                         + Word(alphanums+'.^-+%') * 3)('units')
_nuspectrumcols = Suppress(_nusprange_kw + _spscore_kw + _spsigma_kw)
_nuspectrumvals = (Group(_spectrumbin + _fnums + _fnums)
                   .set_fail_action(trans.fail_spectrum))
_nuspectrum = (Suppress(_nuspectrum_kw)
               + _numdiscbatch
               + _nuspectrumcols
               + Optional(_nuspectrumunits)
               + OneOrMore(_nuspectrumvals, stopOn=_endtable)('spectrum_vals'))
nuspectrumblock = Group(Group(_nuspectrum + integratedres))('nu_spectrum_res')


# ZA spectrum
_zaspectrumunits = Group(Suppress(_units_kw)
                         + Word(alphanums+'.^-+%') * 2)('units')
_zaspectrumcols = Suppress(_zaspid_kw + _spscore_kw + _spsigma_kw)
_zaspectrumbin = (Suppress('(') + _inums + Suppress(',')
                  + _inums + Suppress(')'))
_zaspectrumvals = (Group(_zaspectrumbin + _fnums + _fnums)
                   .set_fail_action(trans.fail_spectrum))
_zaspectrum = (Suppress(_zaspectrum_kw)
               + _numdiscbatch
               + _zaspectrumcols
               + Optional(_zaspectrumunits)
               + OneOrMore(_zaspectrumvals, stopOn=_endtable)('spectrum_vals'))
zaspectrumblock = Group(Group(_zaspectrum + integratedres))('za_spectrum_res')


def _printtoks(toks):
    print(toks)


# Entropy
_boltzmannentropy = (Suppress(_boltzmannentropy_kw)
                     + _fnums('boltzmann_entropy_res'))
_shannonentropy = Suppress(_shannonentropy_kw) + _fnums('shannon_entropy_res')
entropy = _boltzmannentropy + _shannonentropy


# Mesh
_mesh_energyrange = (Group(Suppress(_energyrange_kw + "(in") + Word(alphas)
                           + Suppress('):') + _fnums + Suppress('-') + _fnums)
                     ('mesh_energyrange'))
_mesh_energyintegrated = ((Suppress(_integratedres_name) + Suppress(':'))
                          ('mesh_energyintegrated'))
_mesh_energyline = _mesh_energyrange | _mesh_energyintegrated
_meshres = Group(
    _mesh_energyline
    + Group(original_text_for(
        OneOrMore(Word(printables), stopOn=_endtable)))('mesh_vals')
    + Optional(Group(entropy)('entropy')))
meshblock = Group(OneOrMore(Group(_timestep
                                  + Optional(_score_mesh_unit)
                                  + Group(OneOrMore(_meshres))('meshes')
                                  + Optional(integratedres)))
                  | Group(Group(OneOrMore(_meshres))('meshes')))('mesh_res')


# KIJ matrix
# before keff as KIJ ESTIMATOR used to evaluate KEFF...
# definition de kijdim en variable globale ne fonctionne pas car creee avant
# solution : utiliser forward pour le redefinir !


def _set_kijdim(toks):
    _kijdim = len(toks)
    _kijeigenvec << Group(_fnums * _kijdim)  # pylint: disable=W0106
    _kijmatrix << Group(_fnums * _kijdim)  # pylint: disable=W0106


_kijsum = Group(Optional(Suppress(_kijlefteigenval_kw))
                + Suppress(_kijmkeff_kw) + _fnums
                + Suppress(_kijdomratio_kw) + _fnums)('kijmkeff_res')
_kijeigenval = Group(_fnums + _fnums)
_kijeigenvec = Forward()
_kijmatrix = Forward()
_kijeigenvaltab = (Suppress(_kijeigenval_kw)
                   + (OneOrMore(_kijeigenval).set_parse_action(_set_kijdim))
                   ('kij_eigenval'))
_kijeigenvectab = ((Suppress(_kijeigenvec_kw) + (OneOrMore(_kijeigenvec)))
                   | _kijeigenvecnotprint_kw)('kij_eigenvec')
_kijmatrixtab = (Suppress(_kijmatrix_kw)
                 + ((OneOrMore(_kijmatrix)) | _kijmatrixnotprint_kw)
                 ('kij_matrix'))
kijres = (Group(Suppress(_integratedres_name)
                + _numusedbatch
                + _kijsum
                + _kijeigenvaltab
                + _kijeigenvectab
                + _kijmatrixtab).set_parse_action(trans.convert_kij_result)
          ('kij_res'))


# KIJ SOURCES
_kijsourcesorder = (Suppress(_kijsourcesorder_kw)
                    + Word(alphas)('kij_sources_order') + Suppress(':'))
_kijsourcesval = Group(OneOrMore(_fnums))('kij_sources_vals')
kijsources = (Group(Suppress(_integratedres_name)
                    + _numusedbatch
                    + Suppress(_kijsources_kw)
                    + Optional(_kijsourcesorder)
                    + _kijsourcesval)
              .set_parse_action(trans.convert_kij_sources)('kij_sources'))


# KIJ estimator for keff

def _define_kij_dim(toks):
    LOGGER.debug("KIJ dimension: %d", len(toks))
    _kijdim = len(toks)
    _identifier = (Group(Suppress('(')
                         + _inums + Suppress(',')
                         + _inums + Suppress(',')
                         + _inums + Suppress(')'))
                   | _inums)
    _idline << Group(_identifier * _kijdim)  # pylint: disable=W0106
    _matline << Group(_identifier            # pylint: disable=W0106
                      + Suppress('|')
                      + (_fnums + Suppress('|')) * _kijdim)


def _set_kij_vols(toks):
    _nbvols = int(toks[0])
    _kijlistvol << Group(_inums * _nbvols)  # pylint: disable=W0106


_kijkeffbeg = Word(alphas)('estimator') + Suppress(_estimator_kw) + _minus_line
_kijlistvol = Forward()
_kijfissilevol = ((Suppress(_kijfissilevol_kw) + _inums('nb_fissile_vols'))
                  .set_parse_action(_set_kij_vols)
                  + Suppress(_kijlistfissilevol_kw)
                  + _kijlistvol('list_fissile_vols'))
_kijkeffintro = (Optional(_kijfissilevol)
                 + Suppress(_kijbatchs_kw) + _inums('batchs_kept')
                 + Suppress(_kijmkeff_kw) + _fnums('kij_mkeff'))

_idline = Forward()
_matline = Forward()
_kijkeffev = (OneOrMore(Group(_inums + _fnums))
              .set_parse_action(_define_kij_dim))
_kijkeffevtab = Group(Suppress(_kijkeffevid_kw) + _kijkeffev)('kij_leigenvec')
_defmatrix = _idline + _minus_line + OneOrMore(_matline + _minus_line)
_kijkeffmatrix = (Group(Suppress(_kijkeffmat_kw) + _defmatrix)
                  ('kij_matrix'))
_kijkeffstdmatrix = (Group(Suppress(_kijkeffstddevmat_kw) + _defmatrix)
                     ('kij_stddev_matrix'))
_kijkeffsensibmatrix = (Group(Suppress(_kijkeffsensibilitymat_kw) + _defmatrix)
                        ('kij_sensibility_matrix'))

_kijkeffblock = (Group(_kijkeffbeg
                       + _kijkeffintro
                       + _kijkeffevtab
                       + _kijkeffmatrix
                       + _kijkeffstdmatrix
                       + _kijkeffsensibmatrix)
                 .set_parse_action(trans.convert_kij_keff))


# Keff as generic response
_keffres = Group(Word(alphas) + _fnums + _fnums)
_keffresblock = Group(OneOrMore(_keffres))('res_per_estimator')
_correlationdesc = Suppress(_correlations_kw)
_correlationestim = Group(Word(alphas) + Suppress('<->') + Word(alphas))
_correlationvals = OneOrMore(_fnums
                             | _notconverged_kw)
_correlation = Group(_correlationestim + _correlationvals)
_correlationblock = (Group(_correlationdesc + OneOrMore(_correlation))
                     ('correlation_mat'))
_fullcombestimation = (Group(Suppress(_fullcomb_kw)
                             + ((_fnums + _fnums)
                                | (_notconverged_kw
                                   + Suppress(Optional(_notcvg_exp_kw)))))
                       ('full_comb_estimation'))
_autokeffres = ((_keffresblock + _correlationblock + _fullcombestimation)
                | _notconverged_kw('not_converged'))
_warnkeff = (Suppress(_warning_kw)
             + _warn_combkeff_kw.set_parse_action(' '.join)('warning'))
keffblock = Group(Suppress(_integratedres_name)
                  + _numusedbatch
                  + (_autokeffres | _warnkeff)
                  ).set_parse_action(trans.convert_keff)('keff_res')


# Keff as historical response
_bestresestim = (OneOrMore(Word(alphas), stopOn=_estimator_kw)
                 .set_parse_action(' '.join)('keff_estimator')
                 + Suppress(_estimator_kw))
_bestresdiscbatch = (Suppress(_bestresdiscbatchs_kw)
                     + _inums('best_disc_batchs')
                     + Suppress("batches"))
_bestkeff = (Suppress(Keyword("keff") + '=') + _fnums('keff')
             + Suppress(Keyword("sigma") + '=') + _fnums('sigma')
             + Suppress(Keyword("sigma%") + '=') + _fnums('sigma%'))
_equivkeff = Suppress(_equivkeff_kw) + _fnums('equivalent_keff')
_parna_lh = (Suppress(Keyword("mean") + '=') + _fnums('mean')
             + Suppress(Keyword("lambda") + '=') + _fnums('lambda')
             + Suppress(Keyword("sigma") + '=') + _fnums('sigma')
             + Suppress(Keyword("sigma%") + '=') + _fnums('sigma%'))
_parna_proba = (Suppress(Keyword("proba") + ':') + _fnums('proba')
                + Suppress(Keyword("lower") + '=') + _fnums('lower')
                + Suppress(Keyword("upper") + '=') + _fnums('upper')
                + Suppress(Keyword("length") + '=') + _fnums('length'))
_parna_inv_cdf = Suppress(_parna_inv_cdf_kw + ":"
                          + _inums('iterations')
                          + Keyword('iterations'))
_parna_likelihood = (Suppress(_parna_likelihood_kw) + _parna_lh
                     + Group(_parna_proba + _parna_inv_cdf)('3sigma')
                     + Group(_parna_proba + _parna_inv_cdf)('2.57sigma'))
_parna_block = (Group(_parna_likelihood)('parna'))
_bestkeffpestim = (Group(_notconverged_kw('not_converged')
                         | Group(Optional(_bestresdiscbatch)
                                 + _numusedbatch
                                 + _bestkeff
                                 + Optional(_equivkeff)
                                 + Optional(_parna_block)
                                 )('keff_auto'))
                   ('results'))
_bestreskeff = Group(_bestresestim + _minus_line + _bestkeffpestim)
_warnfixedsources = Group(Suppress(_warning_kw) + _minus_line
                          + _warn_fixsourcekeff_kw('warning'))
_bestresblock = OneOrMore(_bestreskeff, stopOn="KIJ")
autokeffblock = Group(Group(Optional(_warnfixedsources)
                            + _bestresblock
                            + Optional(_kijkeffblock))
                      .set_parse_action(trans.convert_keff_auto))('keff_auto')


# MED files
medfile = (Suppress((_creationmedfile_kw | _creationfile_kw) + ':')
           + Word(alphanums+'/_.')('med_file')
           + Suppress(_medmeshid_kw + Word(alphanums+'_.')))


# Greenbands exploitation
_gbspectrumstep = Suppress(_gbspectrumstep_kw) + _inums
_gbenergymin = Suppress(_gbenergymin_kw) + _fnums
_gbenergymax = Suppress(_gbenergymax_kw) + _fnums
_gbstepdesc = Group(_gbspectrumstep + _minus_line
                    + _gbenergymin + _gbenergymax)('gb_step_desc')
_gbtabulation = (Suppress(_gbsourcetab_kw)
                 + Group(Suppress('u =') + _inums
                         + Suppress(', v =') + _inums
                         + Suppress(', w =') + _inums))
_gbsource = Group(Suppress(_gbsourcenum_kw)
                  + _inums
                  + (_minus_line | (_gbtabulation + _minus_line)))('gb_source')
_gbrespersource = Group(_gbsource + spectrumblock)
_gbstep = Group(_gbstepdesc + Group(OneOrMore(_gbrespersource))('gb_step_res'))
gbblock = Group(OneOrMore(_gbstep))('green_bands_res')


# Scores ordered by nuclei and precursor families, IFP outputs
_generic_score = Suppress(':') + _fnums + _fnums
# Nuclei order alone
_scorepernucleus = Group(Word(alphanums) + _generic_score)
_nucleiorder = (Suppress(_nucleiorder_kw)
                + ZeroOrMore(_scorepernucleus)('score_per_nucleus'))
# Families order alone
_scoreperfamily = Group(Suppress("i =") + _inums + _generic_score)
_familyorder = (Suppress(_familiesorder_kw)
                + ZeroOrMore(_scoreperfamily)('score_per_family'))
# Nuclei and families order
_nucleusid = Suppress(_nucleus_kw) + Word(alphanums) + Suppress('.')
_nucleusfam = Group(_nucleusid + _familyorder)
_nuclfamorder = (Suppress(_nucleifamilyorder_kw)
                 + ZeroOrMore(_nucleusfam)('score_per_nucleus_family'))
# Perturbation index order
_scoreperpertuind = Group(Suppress("i =") + _inums + _generic_score)
_perturborder = (Suppress(_perturborder_kw)
                 + ZeroOrMore(_scoreperpertuind)('score_per_perturbation'))
# Convergence statistics
_cvgline = Group(Suppress("L =") + _inums + _generic_score)
_cvgstat = (Suppress(_cvgstat_kw)
            + ZeroOrMore(_cvgline)('score_per_length'))
# results from adjoint calculation
adjointres = (Group(Group(Suppress(_integratedres_name)
                          + _numusedbatch
                          + Group(_nuclfamorder
                                  | _nucleiorder
                                  | _familyorder
                                  | _perturborder
                                  | _cvgstat)('adj_res')
                          + Optional(_unitsres)
                          ).set_parse_action(trans.convert_generic_adjoint))
              )('adjoint')
# Convergence statistics
_kingcritline = Group(Suppress("CRITICALITY SOURCE") + _fnums + _fnums)
_kingtimestepline = Group(Suppress("END OF TIME STEP ") + Suppress(_inums)
                          + _fnums + _fnums)
_kingstat = Group(_kingcritline
                  + ZeroOrMore(_kingtimestepline))('kin_generic_res')
kingres = (Group(Suppress(_integratedres_name)
                 + _numusedbatch
                 + _kingstat
                 + Optional(_unitsres)('units')
                 ).set_parse_action(trans.convert_generic_kinetic)
           )('kinetic_generic')
# sensitivities
_sensitivityorder = (Suppress(_sensitivitytypeorder_kw)
                     + OneOrMore(Word(alphas + '_,()'),
                                 stopOn=_sensitivityindexorder_kw)
                     .set_parse_action(' '.join)
                     + Suppress(_sensitivityindexorder_kw))
_sensitivityreactionrate_ratio = Group(Suppress(_sensitivityreactionrate_kw)
                                       + _fnums('score') + _fnums('sigma'))
_sensitivity_type = (OneOrMore(Word(alphas.upper()), stopOn=_sensitivity_kw)
                     .set_parse_action(' '.join)
                     + Suppress(_sensitivity_kw))('typeI')
_sensitivity_index = Group(
    Suppress("i =") + _inums('sensitivity_index') + Suppress(';')
    + Suppress("NUCLEUS :") + Word(alphanums + '_')('sensitivity_nucleus')
    + Suppress(',') + Suppress("TYPE :")
    + Word(alphanums.upper() + '_() ')('sensitivity_reaction'))
_sensitivity_dircos = Group(Suppress(_sensitivity_dircos_kw)
                            + _fnums*2)('direction_cosine')
_sensitivity_energyinc = Group(Suppress(_sensitivity_incenergy_kw)
                               + _fnums*2)('energy_incident')
_sensitivity_cols = (Keyword("E min") + Keyword("E max")
                     + Keyword("S(E)") + Keyword("sigma"))
_sensitivity_vals = Group(_fnums*4)
_sensitivity_energyint = Group(Suppress(_sensitivity_energyint_kw)
                               + _integratedres)
_sensitivity_res = Group(_sensitivity_index('charac')
                         + Group(OneOrMore(Group(
                             ZeroOrMore(_sensitivity_dircos
                                        | _sensitivity_energyinc)
                             + Suppress(_sensitivity_cols)
                             + OneOrMore(_sensitivity_vals)('values'))))
                         ('vals')
                         + _sensitivity_energyint('energy_integrated'))
_sensitivity = (Suppress(_sensitivityorder)
                + ZeroOrMore(Group(_sensitivityreactionrate_ratio
                                   ('reaction_rate')))
                + (OneOrMore(Group(_sensitivity_type('sensitivity_type')
                                   + OneOrMore(_sensitivity_res)('res')))))
sensitivityres = Group(Group(Optional(Suppress(_integratedres_name))
                             + _numusedbatch
                             + Group(_sensitivity)('sensit_res')
                             + Optional(_unitsres)('units'))
                       .set_parse_action(trans.convert_sensitivities)
                       )('sensitivity')


def _rename_norm_kw():
    '''Transform RESULTS ARE NORMALIZED keyword in int (lighter)'''
    return 1


def _define_ifp_adj_table_dim(toks):
    '''Define the format of the IFP adjoint criticality table result:
    coordinates (Vol or space coordinates) are followed by energy.
    '''
    _tabdim = len(toks)
    if "Vol" not in toks[0]:
        _ifpadjbinval << _fnums * 2 * _tabdim  # pylint: disable=W0104
    else:
        # in Vol case: int to identify volume, then E (min | max)
        _ifpadjbinval << _inums + _fnums * 2  # pylint: disable=W0104


# IFP adjoint criticality edition
_adjcrit_ed_intro = _star_line + Suppress(_ifpadjcriticality_kw) + _star_line
_ifpadjcrit_intro = Group(Word(alphas+'_')('ifp_response')
                          + _scorename
                          + Suppress(_ifpadjcyclelength_kw)
                          + _inums('ifp_cycle_length')
                          + (Optional(_ifpadjnormalizedres_kw
                                      .set_parse_action(_rename_norm_kw))
                             ('normalized'))
                          + _star_line)
_ifpadjbinval = Forward()
_ifpadjcoordinate = Word(alphas) + Suppress(_ifpadjminmax_kw)
_ifpadjcoordinates = ((Optional(_ifpadjvol_kw) + OneOrMore(_ifpadjcoordinate))
                      .set_parse_action(_define_ifp_adj_table_dim))
_ifpadjcolumns = _ifpadjcoordinates + _ifpadjscore_kw + _spsigma_kw
_ifpadjline = Group(_ifpadjbinval + _fnums + _fnums)
_ifpadjvalues = OneOrMore(_ifpadjline)
_adjcritblock = Group(_ifpadjcrit_intro('ifp_adjoint_criticality_intro')
                      + _ifpadjcolumns('columns')
                      + _ifpadjvalues('values')
                      + _star_line)
ifpadjointcriticality = (
    Group((_adjcrit_ed_intro + OneOrMore(_adjcritblock))
          .set_parse_action(trans.convert_ifp_adj_crit_ed))
    ('ifp_adjoint_crit_edition'))

# Spherical harmonics results
_shr_bins = Group(Suppress(_shr_nb_space_bins_kw) + _inums('nb_sbins')
                  + Suppress(_shr_nb_u_space_bins_kw) + _inums('nb_ubins')
                  + Suppress(_shr_nb_v_space_bins_kw) + _inums('nb_vbins')
                  + Suppress(_shr_nb_w_space_bins_kw) + _inums('nb_wbins')
                  + Suppress(_shr_nb_energy_bins_kw) + _inums('nb_ebins')
                  + Suppress(_shr_nb_inc_energy_bins_kw) + _inums('nb_iebins')
                  + Suppress(_shr_lmax_kw) + _inums('lmax'))('nb_bins')
_shrscore = (Suppress(_shr_score_name_kw)
             + Group(OneOrMore(Word(printables), stopOn=LineEnd())
             .set_parse_action(' '.join))('score_name'))
_shrspace = Suppress(_shr_space_kw) + Group(_inums + _inums + _inums)('space')
_shrenergy = Suppress(_shr_energy_kw) + Group(_fnums + _fnums)('energy')
_shrincenergy = (Suppress(_shr_incident_energy_kw)
                 + Group(_fnums + _fnums)('incident_energy'))
_shrcols = Suppress(_shr_l_kw + _shr_m_kw + _shr_score_kw + _shr_sigma_kw)
_shrvals = Group(_inums + _inums + _fnums + _fnums)
_shrtable = _shrenergy + _shrcols + OneOrMore(_shrvals)('values')
_shr_ietable = Optional(_shrincenergy) + OneOrMore(Group(_shrtable))('vpie')
_shr_spacetable = _shrspace + OneOrMore(Group(_shr_ietable))('vpspace')
_shr_scoretable = _shrscore + OneOrMore(Group(_shr_spacetable))('score')
shrblock = Group(_shr_bins
                 + Group(OneOrMore(Group(_shr_scoretable)))('res')
                 + Suppress(_star_line)
                 ).set_parse_action(trans.convert_shr)('spherical_harmonics')

# Perturbations
_perturank = Suppress(_perturank_kw) + _inums('perturbation_rank')
_pertumethod = (Suppress(_pertumethod_kw)
                + OneOrMore(Word(alphas), stopOn=LineEnd())
                .set_parse_action(' '.join)('perturbation_method'))
_pertuorder = Suppress(_pertuorder_kw) + _inums('perturbation_order')
_pertutype = Suppress(_pertutype_kw) + Word(alphas)('perturbation_type')
_pertucompo = (Suppress(_pertucompo_kw)
               + Word(alphanums+'_')('perturbation_composition'))
pertu_desc = (Group(Suppress(_perturbation_kw)
                    + _perturank
                    + _pertumethod
                    + Optional(_pertuorder)
                    + _pertutype
                    + _pertucompo)('perturbation_desc'))

# Uncertainties (linked to perturbations ?)
_uncertcols = Suppress(_uncertgp_kw + _uncertsig2_kw
                       + _uncertmean_kw + _uncertsig_kw + _uncertfisher_kw)
_uncertbin = _fnums + Suppress('-') + _fnums
_uncertvals = Group(_uncertbin + _fnums * 4)
_uncertspectrum = (Suppress(_uncertres_kw)
                   + _numdiscbatch
                   + _uncertcols
                   + _minus_line
                   + OneOrMore(_uncertvals, stopOn=_endtable)
                   ('spectrum_vals'))
uncertblock = Group(OneOrMore
                    (Group(OneOrMore(_timestep | _muangzone | _phiangzone)
                           + _uncertspectrum))
                    | Group(_uncertspectrum))('uncert_spectrum_res')

_uncertintegres = (_fnums('sigma2(means)')
                   + _fnums('mean(sigma_n2)')
                   + _fnums('sigma(sigma_n2)')
                   + _fnums('fisher test'))
_uncertintegfullres = _numusedbatch + _uncertintegres
uncertintegblock = Group(Suppress(_uncertintegres_kw)
                         + _numdiscbatch
                         + _uncertintegfullres)('uncert_integrated_res')


# Creation jdds
_nbpartline = Group(Suppress("FILE") + _inums
                    + Suppress(':') + _inums
                    + Suppress("particles"))
contribpartblock = (Group(Suppress(_nbcontribpart_kw)
                          + _minus_line
                          + OneOrMore(_nbpartline)
                          + Suppress(_endcontribpart_kw))
                    ('contributing_particles'))


# Score block
scoreblock = (Group(scoredesc + (OneOrMore(vovspectrumblock
                                           | spectrumblock
                                           | nuspectrumblock
                                           | zaspectrumblock
                                           | meshblock
                                           | medfile
                                           | integratedres
                                           | bestres
                                           | uncertblock
                                           | uncertintegblock
                                           | gbblock
                                           | corresptable)))
              .set_parse_action(trans.convert_score))

listscoreblock = (Group(OneOrMore(scoreblock)
                        .set_parse_action(trans.index_elements('score_index')))
                  ('score_res'))

# Response block
responseblock = Group(keffblock
                      | kijres
                      | kijsources
                      | adjointres
                      | kingres
                      | sensitivityres
                      | genericscoreblock
                      | shrblock
                      | listscoreblock)('results')

response = (Group(_star_line
                  + respintro
                  + _star_line
                  + responseblock)
            .set_parse_action(trans.finalize_response_dict))

listresponses = Group(OneOrMore(response).set_parse_action(
    compose2(trans.extract_all_metadata,
             trans.index_elements('response_index'))))('list_responses')

perturbation = (OneOrMore(Group(pertu_desc + listresponses)
                          .set_parse_action(trans.propagate_all_metadata))
                .set_parse_action(trans.index_elements('perturbation_index'))
                ('perturbation'))


################################
#         DEBUG PARSER         #
################################

# debug grammar, to be used with parse_debug (only for parsing development)
t4debug_gram = (OneOrMore((intro
                           + OneOrMore(listresponses | ifpadjointcriticality
                                       | autokeffblock | perturbation
                                       | Suppress(contribpartblock)
                                       | OneOrMore(runtime)))
                          .set_parse_action(trans.to_final_dict))
                .set_parse_action(dump_in_logger)
                | intro + OneOrMore(runtime)
                ).set_fail_action(trans.fail_parsing)


################################
#        GENERAL PARSER        #
################################

t4gram = (OneOrMore((intro
                     + ZeroOrMore(listresponses | ifpadjointcriticality
                                  | autokeffblock | perturbation
                                  | Suppress(contribpartblock))
                     + runtime)
                    .set_parse_action(trans.to_final_dict))
          .set_parse_action(dump_in_logger)
          ).set_fail_action(trans.fail_parsing)



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

r'''This module is designed to pick the required results from an Apollo3
standard HDF5 output file. This means that the users should already know which
result they want to extract. Some navigation methods exist, but they are not
optimised for speed of access.

One possibility is to first explore the output file interactively with a
:class:`~.hdf5_reader.Reader`, and then to write the efficient code using
:class:`Picker`.


Picking results with :class:`Picker`
------------------------------------

.. code-block:: python

    from valjean.eponine.apollo3.hdf5_picker import Picker
    ap3p = Picker('MON_FICHIER.hdf')
    rate_ds = ap3p.pick_standard_value(output='OUTPUT_FOLDER',
                                       zone='ZONE_FOLDER',
                                       result_name='REQUIRED_RESULT')

For example to get the k\ :sub:`eff`, in ``output_1`` under ``totaloutput``:

.. code-block:: python

    keff = ap3p.pick_standard_value(output='output_1', zone='totaloutput',
                                    result_name='KEFF')

The result will be a :class:`~valjean.eponine.dataset.Dataset`. The error is
not available and it is set to ``numpy.nan`` by default. It can be changed
using the ``error_value`` to the :class:`Picker` constructor.

In the case of a *rates* file, bins are set to groups index in most cases. When
the quantity is also calculated on anisotropies (e.g. diffusion) the bins are
``('anisotropies', 'groups')``. The surfacic flux (``'SURFFLUX'``) bins are
``('groups', 'surfaces')`` and the current (``'CURRENT'``) ones are
``('groups', 'surfaces', 'direction')``.

Results per isotopes can be obtained with an additional keyword argument, for
example for :sup:`10`\ B absorption in ``output_0`` from zone ``5``:

.. code-block:: python

    b10abs = ap3p.pick_standard_value(output='output_0', zone='5',
                                      result_name='Absorption', isotope='B10')

Macroscopic rates are obtained with ``isotope='macro'``.

Local values can also be retrived using the
:meth:`Picker.pick_user_value` method:

* if ``'LOCALNAME'`` and ``'LOCALVALUE'`` belong to zone names:

    .. code-block:: python

        locv = ap3p.pick_user_value(output='output', zone=None,
                                    result_name='LOCAL_VALUE_NAME')

* if ``'LOCALNAME'`` and ``'LOCALVALUE'`` belong to result names in a zone:

    .. code-block:: python

        locv = ap3p.pick_user_value(output='output', zone='totaloutput',
                                    result_name='LOCAL_VALUE_NAME')

* if ``'localvalue'`` belongs to zone names:

    .. code-block:: python

        locv = ap3p.pick_user_value(output='output', zone='localvalue',
                                    result_name='LOCAL_VALUE_NAME')

In both cases, it might be useful to check the zones and the local names first.


Quick exploration with :class:`Picker`
--------------------------------------

Some helpers exist if needed:

* :meth:`Picker.outputs`: list of output folders
* :meth:`Picker.geometry`: list of geometries stored in a given output folder
* :meth:`Picker.nb_groups`: number of energy groups available in a given output
  folder
* :meth:`Picker.zones`: list of zones in a given output folder
* :meth:`Picker.isotopes`: list of isotopes in a given zone of a given output
  folder
* :meth:`Picker.results`: list of available results in a given zone of a given
  output folder, isotope can be precised if needed
* :meth:`Picker.nb_anisotropies`: number of anisotropies available for a given
  result for the considered isotope in its zone and output
* :meth:`Picker.local_names`: list of names of the local values stored in a
  given zone from a given output folder
'''

import logging
from functools import lru_cache
from collections import OrderedDict
import h5py
import numpy as np

from ...chrono import Chrono
from ..dataset import Dataset
from ...cosette.rlist import RList


LOGGER = logging.getLogger(__name__)


class Picker:
    '''Pick selected values from an Apollo3 HDF5 file.'''

    CACHE_SIZE = 512

    def __init__(self, fname, error_value=np.nan):
        '''Initialize the :class:`Picker` object and read the HDF5 file.

        :param str fname: path to the Apollo3 HDF5 ouput file
        :param error_value: value to give to the error (default: ``numpy.nan``)
        :type error: int, float
        '''
        LOGGER.info('Reading %s', fname)
        self.error_value = error_value
        with Chrono() as chrono:
            self.hfile = h5py.File(fname, 'r')
        LOGGER.info('HDF5 loading done in %f s', chrono)

    def close(self):
        '''Close HDF5 file.'''
        self.hfile.close()

    def geometry_from_geomid(self, *, geometry):
        '''Return geometry as a dictionary of zone names and volumes.

        :param str geometry: geometry name
        '''
        LOGGER.debug("In geometry_from_geomid")
        ageoms = {
            name.decode('UTF-8').strip(): vol
            for name, vol in zip(self.hfile['geometry'][geometry]['ZONENAME'],
                                 self.hfile['geometry'][geometry]['VOLUME'])}
        return ageoms

    def geometry(self, *, output):
        '''Return geometry characteristics (zones and their volumes) from the
        output name.

        :param str output: output in which zones will be found
        :rtype: dict
        '''
        return self.geometry_from_geomid(
            geometry=self.hfile['info'][output]['GEOMID'][...][0])

    def nb_groups(self, *, output):
        '''Return the number of groups for the `output`.

        :param str output: output in which zones will be found
        :rtype: int
        '''
        return self.hfile['info'][output]['NG'][...][0]

    def outputs(self):
        '''Return list of available outputs.

        :rtype: list(str)
        '''
        return [f for f in self.hfile if 'output' in f]

    def zones(self, *, output):
        '''Return list of available zones in `output`.

        :param str output: output in which zones will be found
        :rtype: list(str)
        '''
        return list(self.hfile[output])

    @lru_cache(maxsize=CACHE_SIZE)
    def isotopes(self, *, output, zone):
        '''Return list of available isotopes in `zone` from `output`.

        :param str output: output folder considered
        :param str zone: zone from output folder
        :rtype: list(str)
        '''
        if zone == 'totaloutput':
            return RList([], key=lambda x: x)
        if self.hfile[output][zone]['NISOT'][...][0] == 0:
            LOGGER.debug('No isotopes available in %s from %s', zone, output)
            if 'macro' in self.hfile[output][zone]:
                return RList(['macro'], key=lambda x: x)
            return RList([], key=lambda x: x)
        return RList([n.decode('UTF-8').strip()
                      for n in self.hfile[output][zone]['ISOTOPE']]
                     + ['macro'], key=lambda x: x)

    def results(self, *, output, zone, isotope=None):
        '''Return available results for the given configuration.

        The `'info'` group is not considered as a result (group containing the
        number of anisotropies), so it is omitted from of the results list.

        :param str output: output folder considered
        :param str zone: zone from the output folder
        :param isotope: isotope considered for results (`'macro'` is
            counted as an isotope, if no isotope is given global results are
            given like flux or keff)
        :type isotope: str or None
        :rtype: list(str)
        '''
        if not isotope:
            lisot = set(self.isotopes(output=output, zone=zone))
            exkeys = lisot | {'NISOT', 'ISOTOPE', 'CONCEN', 'macro'}
            return [k for k in self.hfile[output][zone] if k not in exkeys]
        lres = [r for r in self.hfile[output][zone][isotope] if r != 'info']
        if isotope != 'macro':
            lres.append('concentration')
        return lres

    def nb_anisotropies(self, *, output, zone, isotope, result):
        '''Return number of anisotropies for the given configuration.

        :param str output: output folder considered
        :param str zone: zone from the output folder
        :param isotope: isotope considered for results (`'macro'` is
            counted as an isotope, if no isotope is given global results are
            given like flux or keff)
        :type isotope: str or None
        :param str result: result considered
        :rtype: list(str)
        '''
        isores = self.hfile[output][zone][isotope]
        if 'info' not in isores:
            return None
        if result in isores['info']:
            return isores['info'][result]['nbAnisotropy'][...][0]
        if 'nbAnisotropy' in isores['info']:
            return isores['info']['nbAnisotropy'][...][0]
        return 1

    def _make_dataset(self, data, name, bins=None):
        '''Build dataset from data.

        Data can be a scalar or an array.

        :param data: data to transform in :class:`Dataset`
        :type data: np.generic, np.ndarray
        :param int ngroups: number of groups
        :param int naniso: number of anisotropies
        :rtype: Dataset
        '''
        error = np.full_like(
            data, self.error_value,
            dtype=(float if np.isnan(self.error_value) else data.dtype))
        if not bins:
            return Dataset(value=data, error=error, what=name.lower())
        if 'anisotropies' in bins:
            nb_groups = len(bins['groups'])
            nb_aniso = len(bins['anisotropies'])
            return Dataset(value=data.reshape(nb_aniso, nb_groups),
                           error=error.reshape(nb_aniso, nb_groups),
                           what=name.lower(), bins=bins)
        if 'incident neutron groups' in bins:
            nb_groups = len(bins['groups'])
            nb_incg = len(bins['incident neutron groups'])
            return Dataset(value=data.reshape(nb_incg, nb_groups),
                           error=error.reshape(nb_incg, nb_groups),
                           what=name.lower(), bins=bins)
        return Dataset(value=data, error=error, what=name.lower(), bins=bins)

    @staticmethod
    def _make_bins(data_size, name, ngroups, naniso=None, nsurf=None):
        '''Make bins for the given data.

        :param int data_size: data size
        :param str name: data name
        :param int ngroups: number of energy groups
        :param int naniso: number of anisotropies
        :param int nsurf: number of surfaces
        :rtype: collections.OrderedDict
        '''
        if ngroups == 0:
            return None
        if ngroups == data_size:  # and (naniso is None or naniso == 1):
            return OrderedDict([('groups', np.arange(ngroups))])
        if name == 'SURFFLUX':
            return OrderedDict([('groups', np.arange(ngroups)),
                                ('surfaces', np.arange(nsurf))])
        if name == 'CURRENT':
            return OrderedDict([
                ('groups', np.arange(ngroups)),
                ('surfaces', np.arange(nsurf)),
                ('direction', np.array(['incoming', 'leaving']))])
        if name == 'MultigroupSpectrum':
            incgrps = data_size // ngroups
            return OrderedDict([
                ('incident neutron groups', np.arange(incgrps)),
                ('groups', np.arange(ngroups))])
        if ngroups * naniso != data_size:
            raise PickerException(
                f"Dataset size ({data_size}) not corresponding to "
                f"NG*Nanisotropies {ngroups}*{naniso} for {name}")
        return OrderedDict([('anisotropies', np.arange(naniso)),
                            ('groups', np.arange(ngroups))])

    def pick_standard_value(self, *, output, zone, result_name, isotope=None):
        '''Pick a result according to required parameters from a standard file
        (like keff, flux or rate).'''
        if result_name in ('KEFF', 'KINF', 'NSURF'):
            return self._make_dataset(
                self.hfile[output][zone][result_name][...][0], result_name)
        if not isotope:
            data = self.hfile[output][zone][result_name][...]
            nsurf = (self.hfile[output][zone]['NSURF'][...][0]
                     if 'NSURF' in self.hfile[output][zone] else None)
            bins = self._make_bins(data.size, result_name,
                                   self.nb_groups(output=output), nsurf=nsurf)
            return self._make_dataset(data, result_name, bins=bins)
        if result_name == 'concentration':
            iisot = self.isotopes(output=output, zone=zone).index(isotope)
            return self._make_dataset(
                self.hfile[output][zone]['CONCEN'][iisot], 'concentration')
        data = self.hfile[output][zone][isotope][result_name][...]
        bins = self._make_bins(
            data.size, result_name, self.nb_groups(output=output),
            self.nb_anisotropies(output=output, zone=zone, isotope=isotope,
                                 result=result_name))
        return self._make_dataset(data, result_name, bins)

    @lru_cache(maxsize=CACHE_SIZE)
    def local_names(self, *, output, zone):
        '''Return the list of available local value names.

        .. note::

            The most common zones seem to be in this case: `'totaloutput'`
            (especially in rates files), `None` or `'localvalue'` in core
            files.

        :param str output: output folder considered
        :param zone: zone from the output folder
        :type zone: str or None
        :rtype: list(str)
        '''
        if not zone:
            return RList([n.decode('UTF-8').strip()
                          for n in self.hfile[output]['LOCALNAME']],
                         key=lambda x: x)
        return RList([n.decode('UTF-8').strip()
                      for n in self.hfile[output][zone]['LOCALNAME']],
                     key=lambda x: x)

    def pick_value_from_index(self, *, output, result_index, zone, name):
        '''Pick a user value result from required parameters.

        :param str output: output folder considered
        :param int result_index: index of the required result
        :param str zone: zone from the output folder
        :rtype: Dataset
        '''
        if zone is None:
            return self._make_dataset(
                self.hfile[output]['LOCALVALUE'][result_index][...], name)
        return self._make_dataset(
            self.hfile[output][zone]['LOCALVALUE'][result_index][...], name)

    def pick_user_value(self, *, output, result_name, zone):
        '''Pick a user value result from required parameters (under local value
        in the file).

        :param str output: output folder considered
        :param str result_name: name of the required result
        :param str zone: zone from the output folder
        :rtype: Dataset
        '''
        if zone == 'localvalue':
            return self._make_dataset(
                self.hfile[output][zone][result_name][...], result_name)
        ilocal = self.local_names(output=output, zone=zone).index(result_name)
        return self.pick_value_from_index(output=output, result_index=ilocal,
                                          zone=zone, name=result_name)


class PickerException(Exception):
    '''An error that may be raised by the :class:`Picker` class.'''



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

'''Top-level module for the valjean package.'''

import logging
import sys
from contextlib import contextmanager
from pkg_resources import get_distribution, DistributionNotFound


try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    __version__ = 'unknown'

logging.NOTE = 15
logging.addLevelName(logging.NOTE, "NOTE")


def note(self, message, *args, **kwargs):
    """New results level function"""
    if self.isEnabledFor(logging.NOTE):
        # pylint: disable=protected-access
        self._log(logging.NOTE, message, args, **kwargs)


logging.Logger.note = note


class ValjeanFormatter:
    '''Custom formatter for log messages.

    This formatter class uses a different format for the ``INFO`` message
    level.
    '''

    def __init__(self, fmt):
        '''Initialize the formatter.

        :param str note_fmt: The format string for the ``NOTE`` level.
        :param str fmt: The format string for the other message levels.
        '''
        self.formatter = logging.Formatter(fmt)

    def format(self, record):
        '''Format the given record.'''
        return self.formatter.format(record)


def _configure_logger(logger, level=logging.INFO):
    _formatter = ValjeanFormatter(LOG_CONSOLE_FORMAT)
    _handler = logging.StreamHandler(sys.stdout)
    _handler.setFormatter(_formatter)
    _handler.setLevel(level)
    logger.addHandler(_handler)
    logger.setLevel(level)


LOGGER = logging.getLogger(__name__)
LOG_CONSOLE_FORMAT = '%(levelname)9.9s %(module)12.12s: %(message)s'
LOG_FILE_FORMAT = ('%(levelname)9.9s (%(module)10.10s/%(funcName)12.12s) '
                   '%(asctime)19s: %(message)s')
_configure_logger(LOGGER)


def set_log_level(level):
    '''Set the verbosity level for the default logger.'''
    LOGGER.setLevel(level)
    for handler in LOGGER.handlers:
        handler.setLevel(level)


def get_logger(module):
    '''Get logger for the given module'''
    logger = logging.getLogger(module)
    return logger


def set_logger_level(module, level=logging.INFO):
    '''Set the verbosity level for the required module logger.

    :param str module: module name
    :param int level: logger level (logging.DEBUG, logging.NOTE, \
        logging.INFO, logging.WARNING, logging.ERROR)
    '''
    logger = get_logger(module)
    logger.setLevel(level)


@contextmanager
def local_logger_level(module, level):
    '''Set the verbosity level for the required module logger in a context
    case.

    :param str module: module name
    :param int level: logger level (logging.DEBUG, logging.NOTE, \
        logging.INFO, logging.WARNING, logging.ERROR)
    '''
    logger = get_logger(module)
    log_level = (logger.getEffectiveLevel() if module == 'valjean'
                 else logging.NOTSET)
    try:
        logger.setLevel(level)
        for handler in logger.handlers:
            handler.setLevel(level)
        yield
    finally:
        logger.setLevel(log_level)



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

# pylint: disable=trailing-whitespace
# pylint: disable=too-many-lines
'''Common module to structure data in same way for all codes.

.. _numpy indexing: https://numpy.org/doc/stable/user/basics.indexing.html
.. _numpy masked: https://numpy.org/doc/stable/reference/maskedarray.html

.. doctest:: dataset
    :hide:

    >>> # noqa: W291

Dataset definition and initialisation
-------------------------------------

A dataset is composed by 5 members:

  * value: a :obj:`numpy.ndarray` or a :obj:`numpy.generic` (scalar from numpy
    represented like the arrays, with dim, etc)
  * error: an object of same type as value
  * bins: an :obj:`collections.OrderedDict` (optional and named argument)
  * name: name of the dataset (optional and named argument)
  * what: can be used to store the name of the quantity represented by the
    dataset (optional and named argument)

The bins object should have the same dimension as the value, the order matches
the dimensions. If no bins are available it is still possible to use an empty
:obj:`collections.OrderedDict`.

>>> from valjean.eponine.dataset import Dataset
>>> import numpy as np
>>> from collections import OrderedDict
>>> bins = OrderedDict([('e', np.array([1, 2, 3])), ('t', np.arange(5))])
>>> ds1 = Dataset(np.arange(10).reshape(2, 5),
...               np.array([0.3]*10).reshape(2, 5),
...               bins=bins, name='ds1', what='spam')
>>> ds1.name
'ds1'
>>> ds1.what
'spam'
>>> len(bins) == ds1.ndim
True
>>> ds1.error.shape == ds1.value.shape
True
>>> ds1.value.shape == (2, 5)
True
>>> np.array_equal(ds1.value, [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]])
True
>>> np.array_equal(ds1.error,
...                [[0.3, 0.3, 0.3, 0.3, 0.3], [0.3, 0.3, 0.3, 0.3, 0.3]])
True
>>> list(bins.keys())
['e', 't']
>>> np.array_equal(bins['e'], [1, 2, 3])
True
>>> np.array_equal(bins['t'], [0, 1, 2, 3, 4])
True

A new dataset can also be created from an already existing one, using ``copy``.
No matter how the dataset is generated, attributes can be changed afterwards:

>>> nds = ds1.copy()
>>> nds.name = 'egg'
>>> nds.what = ''
>>> print(f'name: ds1={ds1.name!r}, nds={nds.name!r}')
name: ds1='ds1', nds='egg'
>>> print(f'what: ds1={ds1.what!r}, nds={nds.what!r}')
what: ds1='spam', nds=''
>>> np.array_equal(ds1.value, nds.value)
True


Errors are emitted if the arguments do not have the expected type or if the
shapes or dimensions are not consistent:

>>> tds = Dataset([1, 2, 3], [0.5, 0.5, 0.5])
Traceback (most recent call last):
        [...]
TypeError: value does not have the expected type (numpy.ndarray or \
numpy.generic = scalar)

>>> tds = Dataset(np.arange(6).reshape(2, 3), np.arange(6).reshape(3, 2))
Traceback (most recent call last):
        [...]
ValueError: Value and error do not have the same shape

>>> tds = Dataset(np.arange(6).reshape(2, 3),
...               np.array([0.5]*6).reshape(2, 3),
...               bins={'spam': [1, 2], 'egg': [1, 2, 3]})
Traceback (most recent call last):
        [...]
TypeError: bins should be an OrderedDict

>>> tds = Dataset(np.arange(6).reshape(2, 3),
...               np.array([0.5]*6).reshape(2, 3),
...               bins=OrderedDict([('spam', [1, 2])]))
Traceback (most recent call last):
        [...]
ValueError: Number of dimensions of bins does not correspond to number of \
dimensions of value


Squeezing a dataset
-------------------

If there are useless dimensions, the dataset can be squeezed:

>>> vals = np.arange(6).reshape(1, 2, 1, 3)
>>> errs = np.array([0.1]*6).reshape(1, 2, 1, 3)
>>> bins2 = OrderedDict([('bacon', np.array([0, 1])),
...                      ('egg', np.array([0, 2, 4])),
...                      ('sausage', np.array([10, 20])),
...                      ('spam', np.array([-5, 0, 5, 10]))])
>>> ds = Dataset(vals, errs, bins=bins2)
>>> ds.value.shape == (1, 2, 1, 3)
True
>>> len(ds.bins) == 4
True
>>> np.array_equal(ds.value, np.array([[[[0, 1, 2]], [[3, 4, 5]]]]))
True
>>> sds = ds.squeeze()
>>> sds.ndim
2
>>> len(sds.bins) == 2
True
>>> sds.shape
(2, 3)
>>> list(len(x)-1 for x in sds.bins.values())  # edges of bins, so N+1
[2, 3]
>>> list(sds.bins.keys()) == ['egg', 'spam']
True
>>> np.array_equal(sds.value, np.array([[0, 1, 2], [3, 4, 5]]))
True

The dimensions with only one bin are squeezed, the same is done on the bins.

>>> nsds = ds.squeeze()
>>> np.array_equal(nsds.value, sds.value)
True
>>> nsds.bins == sds.bins
True


Operations available on Datasets
--------------------------------

Standard operations, ``(+, -, *, /, [])``, are available for :class:`Dataset`.
Examples of their use are given, including failing cases. Some are used in
various methods but shown only once.

All operations conserve the name of the initial dataset.

The **what** attribute can be used to store the name of the quantity
represented by the dataset. It is updated if it involves another dataset:

* in addition or substraction, if **what** is different it contains both
  separated by the symbol
* in multiplication or division it always contains both separeted by the symbol

Addition and subtraction
^^^^^^^^^^^^^^^^^^^^^^^^
Addition and subtraction are possible between a :class:`Dataset` and a scalar
(:obj:`numpy.generic`), a :obj:`numpy.ndarray` and another :class:`Dataset`.
The operator to use for addition is ``+`` and for subtraction ``-``.

Restrictions in addition or subtraction with a :obj:`numpy.ndarray` are handled
by `NumPy`.

The addition or subtraction of two :class:`Dataset` can be done if

* both values have the same shape (:func:`consistent_datasets`)
* bins conditions (:func:`same_coords`):

    * EITHER the second :class:`Dataset` does not have any bins
    * OR bins are the same, i.e. have the same keys and bins values


Example addition of a scalar value (only on value)
``````````````````````````````````````````````````

    >>> ds1p10 = ds1 + 10
    >>> np.array_equal(ds1p10.value,
    ...                [[10, 11, 12, 13, 14], [15, 16, 17, 18, 19]])
    True
    >>> np.array_equal(ds1p10.error, ds1.error)
    True
    >>> ds1p10.bins == ds1.bins
    True
    >>> ds1p10.what == ds1.what
    True

As expected it 'only' acts on the value, error and bins are unchanged.


Example of subtraction of a :obj:`numpy.ndarray`
````````````````````````````````````````````````

    >>> a = np.array([100]*10).reshape(2, 5)
    >>> ds1.value.shape == a.shape
    True
    >>> ds1ma = ds1 - a
    >>> ds1ma.__class__
    <class 'valjean.eponine.dataset.Dataset'>
    >>> np.array_equal(ds1ma.value, [[-100, -99, -98, -97, -96],
    ...                              [-95, -94, -93, -92, -91]])
    True
    >>> np.array_equal(ds1ma.error, ds1.error)
    True
    >>> ds1ma.bins == ds1.bins
    True
    >>> ds1ma.what == ds1.what
    True

``a`` and ``ds1`` have the same shape to everything is fine.

    >>> b = np.array([100]*10)
    >>> ds1.value.shape == b.shape
    False
    >>> ds1pb = ds1 + b
    Traceback (most recent call last):
        ...
    ValueError: operands could not be broadcast together with shapes (2,5) \
(10,) 


If the shapes are not the same `NumPy` raises an exception (and ds1pb is not
defined).


.. _addition_dsataset:

Example of addition or subtraction of another :class:`Dataset`
``````````````````````````````````````````````````````````````

    >>> ds2 = Dataset(value=np.arange(20, 30).reshape(2, 5),
    ...               error=np.array([0.4]*10).reshape(2, 5),
    ...               bins=bins, name='ds2', what='spam')
    >>> ds1.bins == ds2.bins
    True
    >>> ds1p2 = ds1 + ds2
    >>> ds1p2.name == ds1.name
    True
    >>> ds1p2.what == ds1.what
    True
    >>> np.array_equal(ds1p2.value, [[20, 22, 24, 26, 28],
    ...                              [30, 32, 34, 36, 38]])
    True
    >>> np.array_equal(ds1p2.error, np.array([0.5]*10).reshape(2, 5))
    True

The error is calulated considering both datasets are independent, so
quadratically (:math:`e = \\sqrt{ds1.e^2 + ds2.e^2}`).

Datasets without binning can also be added:

    >>> ds3 = Dataset(value=np.arange(200, 300, 10).reshape(2, 5),
    ...               error=np.array([0.4]*10).reshape(2, 5), name='ds3')
    >>> ds3.bins
    OrderedDict()
    >>> ds1p3 = ds1 + ds3
    >>> ds1p3.name == ds1.name
    True
    >>> ds1p3.what == ds1.what
    False
    >>> ds1p3.what
    'spam+'
    >>> np.array_equal(ds1p3.value, [[200, 211, 222, 233, 244],
    ...                              [255, 266, 277, 288, 299]])
    True
    >>> np.array_equal(ds1p3.error, np.array([0.5]*10).reshape(2, 5))
    True
    >>> same_coords(ds1p3, ds1)
    True

Bins of the dataset on the left are kept.

Like in `NumPy` array addition, values need to have the same shape:

    >>> ds4 = Dataset(np.arange(5), np.array([0.01]*5), name='ds4')
    >>> f"shape ds1 {ds1.value.shape}, ds4 {ds4.value.shape} -> \
comp = {ds1.value.shape == ds4.value.shape}"
    'shape ds1 (2, 5), ds4 (5,) -> comp = False'
    >>> ds1 + ds4
    Traceback (most recent call last):
        [...]
    ValueError: Datasets to add do not have same shape

If bins are given, they need to have the same keys and the same values.

    >>> bins5 = OrderedDict([('E', np.array([1, 2, 3])), ('t', np.arange(5))])
    >>> ds5 = Dataset(np.arange(0, -10, -1).reshape(2, 5),
    ...               np.array([0.01]*10).reshape(2, 5),
    ...               bins=bins5, name='ds5')
    >>> ds1 + ds5
    Traceback (most recent call last):
        [...]
    ValueError: Datasets to add do not have same bin names
    >>> f"bins ds1: {list(ds1.bins.keys())}, bins ds5: {list(ds5.bins.keys())}"
    "bins ds1: ['e', 't'], bins ds5: ['E', 't']"

    >>> bins6 = OrderedDict([('e', np.array([1, 2, 30])), ('t', np.arange(5))])
    >>> ds6 = Dataset(np.arange(0, -10, -1).reshape(2, 5),
    ...               np.array([0.01]*10).reshape(2, 5),
    ...               bins=bins6, name='ds6')
    >>> ds1 - ds6
    Traceback (most recent call last):
        [...]
    ValueError: Datasets to subtract do not have the same bins
    >>> same_coords(ds1, ds6)
    False
    >>> list(ds1.bins.keys()) == list(ds6.bins.keys())
    True
    >>> np.array_equal(ds1.bins['e'], ds6.bins['e'])
    False
    >>> np.array_equal(ds1.bins['t'], ds6.bins['t'])
    True


Multiplication and division
^^^^^^^^^^^^^^^^^^^^^^^^^^^
Multiplication and division are possible between a :class:`Dataset` and a
scalar (:obj:`numpy.generic`), a :obj:`numpy.ndarray` and another
:class:`Dataset`.
The operator to use for multiplication is ``*`` and for division ``/``.

Restrictions in multiplication and division with a :obj:`numpy.ndarray` are
handled by `NumPy`.

The multiplication or division of 2 :class:`Dataset` can be done if

* both values have the same shape (:func:`consistent_datasets`)
* bins conditions (:func:`same_coords`):

    * EITHER the second :class:`Dataset` does not have any bins
    * OR bins are the same, i.e. have the same keys and bins values

Division by zero, nan or inf are handled by `NumPy` and return a
RuntimeWarning from `NumPy` (only in the zero case).


Example multiplication of a scalar value
````````````````````````````````````````

    >>> ds1m10 = ds1 * 10
    >>> ds1m10.__class__
    <class 'valjean.eponine.dataset.Dataset'>
    >>> np.array_equal(ds1m10.value, [[0, 10, 20, 30, 40],
    ...                               [50, 60, 70, 80, 90]])
    True
    >>> np.array_equal(ds1m10.error, np.array([3.]*10).reshape(2, 5))
    True
    >>> same_coords(ds1m10, ds1)
    True

As expected it acts on the value and on the error. Bins are unchanged.


.. _division_by_nparray:

Example of division of a :obj:`numpy.ndarray`
`````````````````````````````````````````````

    >>> ds1da = ds1 / a
    >>> ds1da.name == ds1.name
    True
    >>> ds1da.what == ds1.what
    True
    >>> np.array_equal(ds1da.value, [[0., 0.01, 0.02, 0.03, 0.04],
    ...                              [0.05, 0.06, 0.07, 0.08, 0.09]])
    True
    >>> np.array_equal(ds1da.error, np.array([0.003]*10).reshape(2, 5))
    True
    >>> same_coords(ds1da, ds1)
    True

``a`` and ``ds1`` have the same shape to everything is fine.

    >>> ds1 / b
    Traceback (most recent call last):
        ...
    ValueError: operands could not be broadcast together with shapes (2,5) \
(10,) 


If the shapes are not the same `NumPy` raises an exception.

If the :obj:`numpy.ndarray`: contains ``0``, ``nan`` or ``inf``, `NumPy` deals
with them. It sends a RunningWarning about the division by zero.

    >>> c = np.array([[2., 3., np.nan, 4., 0.], [1., np.inf, 5., 10., 0.]])
    >>> ds1 / c  # doctest: +SKIP
    class: <class 'valjean.gavroche.dataset.Dataset'>, data type: \
<class 'numpy.ndarray'>
            name: ds1, with shape (2, 5),
            value: [[0.         0.33333333        nan 0.75              inf]
     [5.         0.         1.4        0.8               inf]],
            error: [[0.15       0.03333333        nan 0.075             inf]
     [0.3        0.         0.06       0.03              inf]],
            bins: OrderedDict([('e', array([1, 2, 3])), ('t', \
array([0, 1, 2, 3, 4]))])
    >>> # prints *** RuntimeWarning: divide by zero encountered in true_divide


Example of multiplication or division of another :class:`Dataset`
`````````````````````````````````````````````````````````````````

    >>> ds1m2 = ds1 * ds2
    >>> ds1m2  # doctest: +SKIP
    class: <class 'valjean.gavroche.dataset.Dataset'>, data type: \
<class 'numpy.ndarray'>
            shape (2, 5),
            value: [[  0  21  44  69  96]
     [125 156 189 224 261]],
            error: [[6.         6.31268564 6.64830806 7.00357052 7.37563557]
     [7.76208735 8.16088231 8.57029754 8.98888202 9.4154129 ]],
            bins: OrderedDict([('e', array([1, 2, 3])), ('t', \
array([0, 1, 2, 3, 4]))])
    >>> ds1m2.what
    'spam*spam'
    >>> same_coords(ds1m2, ds2)
    True

    >>> ds1o2 = ds1 / ds2
    >>> ds1o2  # doctest: +SKIP
    class: <class 'valjean.gavroche.dataset.Dataset'>, data type: \
<class 'numpy.ndarray'>
            shape (2, 5),
            value: [[0.         0.04761905 0.09090909 0.13043478 0.16666667]
     [0.2        0.23076923 0.25925926 0.28571429 0.31034483]],
            error: [[0.015      0.01431448 0.01373617 0.01323926 0.01280492]
     [0.01241934 0.01207231 0.01175624 0.01146541 0.0111955 ]],
            bins: OrderedDict([('e', array([1, 2, 3])), ('t', \
array([0, 1, 2, 3, 4]))])
    >>> ds1o2.what
    'spam/spam'

In both cases the error is calulated considering both datasets are independent,
so quadratically :math:`e = v*\\sqrt{{(\\frac{ds_1.e}{ds_1.v})}^2 +
{(\\frac{ds_2.e}{ds_2.v})}^2}`.

The same restictions on bins as for addition and subtraction are set for
multiplication and division, same ``AssertError`` are raised, see
:ref:`addition_dsataset`.

About the division by ``0``, ``nan`` or ``inf``, it acts like in the
multiplication or division by a :obj:`numpy.ndarray`, see
:ref:`division_by_nparray` (warnings and ``nan``, ``inf``, etc.)

>>> np.isnan((ds1/ds1).value[0][0])
True
>>> np.isinf(((ds1+1)/ds1).value[0][0])
True


Indexing and slicing
--------------------

It is only possible to get a **slice** of a dataset, getting an Dataset at a
given index is not possible (for dimensions consistency reasons). Requiring a
given index can then be done using a slice.


Getting a subset of the :class:`Dataset`
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Time is the second dimension, to remove first and last bins the usual slice is
``[1:-1]``, the first dimension, energy, is conserved, so its slice is ``[:]``.
The slice to apply is then ``[:, 1:-1]``.

    >>> ds1sltfl = ds1[:, 1:-1]
    >>> ds1sltfl.__class__
    <class 'valjean.eponine.dataset.Dataset'>
    >>> np.array_equal(ds1sltfl.value, [[1, 2, 3], [6, 7, 8]])
    True
    >>> np.array_equal(ds1sltfl.error, np.array([0.3]*6).reshape(2, 3))
    True
    >>> list(ds1sltfl.bins.keys()) == list(ds1.bins.keys())
    True
    >>> np.array_equal(ds1sltfl.bins['e'], ds1.bins['e'])
    True
    >>> np.array_equal(ds1sltfl.bins['t'], ds1.bins['t'])
    False
    >>> np.array_equal(ds1sltfl.bins['t'], [1, 2, 3])
    True
    >>> ds1sltfl.name == ds1.name
    True

Slicing is also applied on bins.


.. warning::
    Requiring a slice when there are not enough elements on the dimension give
    empty arrays.

    For example: removing first and last bin in energy on ``ds1``. The slice is
    ``[1:-1, :]`` in that case, but ``ds1`` has only 2 bins in energy.

    >>> ds1slefl = ds1[1:-1, :]
    >>> ds1slefl.value.shape == (0, 5)
    True
    >>> np.array_equal(ds1slefl.value, np.array([]).reshape(0, 5))
    True
    >>> np.array_equal(ds1slefl.error, np.array([]).reshape(0, 5))
    True
    >>> np.array_equal(ds1slefl.bins['t'], ds1.bins['t'])
    True
    >>> np.array_equal(ds1slefl.bins['e'], ds1.bins['e'])
    False
    >>> np.array_equal(ds1slefl.bins['e'], [2])
    True

    Note that in this case, as bins are in reality the edges of the bins, so we
    have N+1 values in the bins compared to the value/error where we have N
    values. Slicing then give one value in energy bins, so unusable here (it
    would be empty if we have values and centers of bins instead of edges of
    bins).


All dimensions have to be present in the slice
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Use ``:`` for the untouched dimensions. Number of ``,`` has to be dimension -1.

Let's consider a new Dataset, with 4 dimensions:

    >>> bins2 = OrderedDict([('e', np.arange(4)), ('t', np.arange(3)),
    ...                      ('mu', np.arange(3)), ('phi', np.arange(5))])
    >>> ds6 = Dataset(np.arange(48).reshape(3, 2, 2, 4),
    ...               np.array([0.5]*48).reshape(3, 2, 2, 4),
    ...               bins=bins2, name='ds6')
    >>> ds6.value.ndim == 4
    True

To remove first bin on energy dimension and last bin on phi dimension, the
slice to be used is: ``[1:, :, :, :-1]``.

    >>> ds6_1 = ds6[1:, :, :, :-1]
    >>> ds6.value.shape == (3, 2, 2, 4)
    True
    >>> ds6_1.value.shape == (2, 2, 2, 3)
    True
    >>> np.array_equal(ds6_1.value, [[[[16, 17, 18], [20, 21, 22]],
    ...                               [[24, 25, 26], [28, 29, 30]]],
    ...                              [[[32, 33, 34], [36, 37, 38]],
    ...                               [[40, 41, 42], [44, 45, 46]]]])
    True
    >>> np.array_equal(ds6_1.error, np.array([0.5]*24).reshape(2, 2, 2, 3))
    True
    >>> list(ds6_1.bins.keys()) == list(ds6.bins.keys())
    True
    >>> np.array_equal(ds6_1.bins['e'], ds6.bins['e'][1:])
    True
    >>> np.array_equal(ds6_1.bins['t'], ds6.bins['t'])
    True
    >>> np.array_equal(ds6_1.bins['mu'], ds6.bins['mu'])
    True
    >>> np.array_equal(ds6_1.bins['phi'], ds6.bins['phi'][:-1])
    True

If we only want the second bin in time keeping all bins in energy and direction
angles, the slice is ``[:, 1:2, :, :]``.

    >>> ds6_2 = ds6[:, 1:2, :, :]
    >>> ds6_2.value.shape == (3, 1, 2, 4)
    True
    >>> np.array_equal(ds6_2.value, [[[[8,  9, 10, 11], [12, 13, 14, 15]]],
    ...                              [[[24, 25, 26, 27], [28, 29, 30, 31]]],
    ...                              [[[40, 41, 42, 43], [44, 45, 46, 47]]]])
    True
    >>> list(ds6_2.bins.keys()) == list(ds6.bins.keys())
    True
    >>> all(x.size == y+1
    ...     for x, y in zip(ds6_2.bins.values(), ds6_2.value.shape)) == True
    True
    >>> np.array_equal(ds6_2.bins['t'], [1, 2])
    True

Bins are changed accordingly.


.. warning::
    Comparison to `NumPy`: index and ellipsis are other slicing possibilities
    on :obj:`numpy.ndarray` (see `numpy indexing`_ for current version of
    `NumPy`), but they are disabled here to avoid confusions. Errors are raised
    if required.

    >>> ds1[1]
    Traceback (most recent call last):
        [...]
    TypeError: Index can only be a slice or a tuple of slices

    >>> ds6_2e = ds6[:, 1, :, :]
    Traceback (most recent call last):
        [...]
    TypeError: Index can only be a slice or a tuple of slices

    >>> ds6e = ds6[1:, ..., :-1]
    Traceback (most recent call last):
        [...]
    TypeError: Index can only be a slice or a tuple of slices

    It also need to have the same dimension as the value:

    >>> ds6_2e = ds6[:, 1:2]
    Traceback (most recent call last):
        [...]
    ValueError: len(index) should have the same dimension as the value \
numpy.ndarray, i.e. (# ',' = dim-1).
    ':' can be used for a slice (dimension) not affected by the selection.
    Slicing is only possible if ndim == 1

    A single slice is only possible for arrays for dimension 1:

    >>> ds6_2f = ds6[1:2]
    Traceback (most recent call last):
        [...]
    ValueError: len(index) should have the same dimension as the value \
numpy.ndarray, i.e. (# ',' = dim-1).
    ':' can be used for a slice (dimension) not affected by the selection.
    Slicing is only possible if ndim == 1

    >>> ds7 = Dataset(value=np.arange(48), error=np.array([1]*48))
    >>> ds7.ndim
    1
    >>> ds7.shape
    (48,)
    >>> ds7_extract = ds7[20:24]
    >>> np.array_equal(ds7_extract.value, [20, 21, 22, 23])
    True

.. warning::
    Slicing can also only be applied on :obj:`numpy.ndarray`, not on
    :obj:`numpy.generic`:

    >>> ds8 = Dataset(value=100, error=1)
    >>> ds8[0:1]
    Traceback (most recent call last):
        [...]
    TypeError: [] (__getitem__) can only be applied on numpy.ndarrays


Masked datasets
---------------

In some case it can be useful to mask some elements of a dataset. This
functionality is provided by the `numpy masked`_ arrays module. In the case of
datasets, once the mask is given it is applied to both the value and the error.

>>> mask = np.ma.masked_greater(ds1.value, 6).mask
>>> np.array_equal(mask, [[False, False, False, False, False],
...                       [False, False, True, True, True]])
True
>>> mds = ds1.mask(mask)
>>> np.ma.is_masked(ds1.value)
False
>>> np.ma.is_masked(mds.value)
True
>>> np.ma.is_masked(mds.error)
True

Bins and shape are kept.

>>> mds.shape == ds1.shape
True
>>> np.array_equal(mds.bins['e'], ds1.bins['e'])
True
>>> np.array_equal(mds.bins['t'], ds1.bins['t'])
True

The mask is propagated when performing operations on dataset.

>>> np.sum(ds1.value) == 45
True
>>> np.sum(mds.value) == 21
True
>>> sds = ds1 + mds
>>> np.ma.is_masked(sds.value)
True
>>> np.ma.is_masked(sds.error)
True
>>> np.sum(sds.value) == 42
True
'''
# pylint: enable=trailing-whitespace
import logging
from collections import OrderedDict
import numpy as np


LOGGER = logging.getLogger(__name__)


class Dataset:
    '''Common class for data from all codes.

    For the moment, units are not treated (removed).

    .. todo::

        Think about units. Possibility: using a units package from scipy.

    .. todo::

        How to deal with bins of N values (= center of bins)
    '''
    def __init__(self, value, error, *, bins=None, name='', what=''):
        '''Dataset class initialization.

        :param value: array of N dimensions representing the values
        :type value: int or float or numpy.ndarray or numpy.generic
        :param error: array of N dimensions representing the **absolute**
          errors
        :type error: int or float or numpy.ndarray or numpy.generic
        :param bins: bins corresponding to value (named optional parameter)
        :type bins: collections.OrderedDict (str, numpy.ndarray)
        :param str name: name of the dataset (used in test representation)
        :param str what: name of the quantity represented by the dataset
        '''
        if isinstance(value, float):
            value = np.float64(value)
        if isinstance(error, float):
            error = np.float64(error)
        if isinstance(value, int):
            value = np.int64(value)
        if isinstance(error, int):
            error = np.int64(error)
        if not isinstance(value, (np.ndarray, np.generic)):
            raise TypeError("value does not have the expected type "
                            "(numpy.ndarray or numpy.generic = scalar)")
        if not isinstance(error, (np.ndarray, np.generic)):
            raise TypeError("error does not have the expected type "
                            "(numpy.ndarray or numpy.generic = scalar)")
        LOGGER.debug("Type of value: %s", type(value))
        LOGGER.debug("Value charac: ndim = %s, shape = %s, size = %s",
                     value.ndim, value.shape, value.size)
        if value.shape != error.shape:
            raise ValueError("Value and error do not have the same shape")
        if bins is not None:
            if not isinstance(bins, OrderedDict):
                raise TypeError("bins should be an OrderedDict")
            LOGGER.debug("Number of dimensions from bins: %s", len(bins))
            if bins and len(bins) != value.ndim:
                raise ValueError("Number of dimensions of bins does not "
                                 "correspond to number of dimensions of value")
            if bins and any(len(b) not in (s, s+1)
                            for b, s in zip(bins.values(), value.shape)
                            if len(b)):
                _lens = [len(b) for b in bins.values()]
                raise ValueError('Number of bins does not correspond to value '
                                 f'shape, bins={_lens}, shape={value.shape}')
        self.value = value
        self.error = error
        self.bins = bins.copy() if bins is not None else OrderedDict()
        self.name = name
        self.what = what

    def copy(self):
        '''Return a deep copy of `self`.'''
        return Dataset(value=self.value.copy(),
                       error=self.error.copy(),
                       bins=self.bins.copy(),
                       name=self.name, what=self.what)

    def __repr__(self):
        if isinstance(self.value, np.ndarray):
            return (f"class: {self.__class__}, data type: {type(self.value)}\n"
                    f"        shape: {self.value.shape}\n"
                    f"        value: {self.value.squeeze()},\n"
                    f"        error: {self.error.squeeze()},\n"
                    f"        bins: {self.bins},\n"
                    f"        name: {self.name!r}, what: {self.what!r}")
        return (
            f"class: {self.__class__}, data type: {type(self.value)}\n"
            f"value: {self.value:6e}, error: {self.error:6e}, "
            f"bins: {self.bins}\n"
            f"name: {self.name!r}, what: {self.what!r}\n"
        )

    def __str__(self):
        if isinstance(self.value, np.ndarray):
            bins_str = []
            for key, val in self.bins.items():
                str_val = str(val).replace('\n', '')
                bins_str.append(f"{key}: {str_val}")
            return (f"shape: {self.value.shape}, dim: {self.value.ndim}, "
                    f"type: {type(self.value)}, bins: {bins_str}, "
                    f"name: {self.name}, what: {self.what}")

        return (
            f"value: {self.value:6e}, error: {self.error:6e}, "
            f"bins: {self.bins}, type: {type(self.value)},"
            f"name: {self.name}, what: {self.what}"
        )

    def squeeze(self):
        '''Squeeze dataset: remove useless dimensions.

        Squeeze is based on the shape and on the bins dim ??? To confirm...
        First squeeze bins, then arrays.
        Edges, if only one bin are not kept. Example: spectrum with one bin in
        energy (quite common)
        '''
        key_axis = dict(enumerate(self.bins))
        lbins = self.bins.copy()
        for axis, dim in enumerate(self.shape):
            if dim < 2:
                lbins.pop(key_axis[axis])
        return self.__class__(self.value.squeeze(),
                              self.error.squeeze(),
                              bins=lbins,
                              name=self.name,
                              what=self.what)

    @property
    def shape(self):
        '''Return the data shape, as a read-only property.'''
        return self.value.shape

    @property
    def ndim(self):
        '''Return the data dimension, as a read-only property.'''
        return self.value.ndim

    @property
    def size(self):
        '''Return the data size (total number of elements in the array), as a
        read-only property.'''
        return self.value.size

    def data(self):
        '''Generator yielding objects supporting the buffer protocol that (as a
        whole) represent a serialized version of `self`.

        >>> from valjean.fingerprint import fingerprint
        >>> vals = np.arange(10).reshape(2, 5)
        >>> errs = np.array([1]*10).reshape(2, 5)
        >>> ds = Dataset(vals, errs)
        >>> fingerprint(ds)
        'a8411470d7766c543e90f0f38241dc918b9448d1b9d19b0a9b8b6c91f61944d0'
        >>> bins = OrderedDict([('bacon', np.arange(3)),
        ...                     ('egg', np.arange(5))])
        >>> ds = Dataset(vals, errs, bins=bins)
        >>> fingerprint(ds)
        'dfec4e6ac118d30b7a83cfc500fefaec94f93a76a024d7550732cd08fbfb0fee'
        >>> ds = Dataset(vals, errs, bins=bins, name='name')
        >>> fingerprint(ds)
        '441623c096bf917414013ebbfe345c66124f9a8dfbbd7051cf733ea20d7b651a'
        >>> ds = Dataset(vals, errs, bins=bins, name='name', what='what!')
        >>> fingerprint(ds)
        '9a42c91381d696c1af651665f84efc43efa14d9c471c4ac851239ed0779a48a0'
        '''
        yield self.__class__.__name__.encode('utf-8')
        yield np.require(self.value, requirements='C').data.cast('b')
        yield np.require(self.error, requirements='C').data.cast('b')
        if self.bins is not None:
            for key, val in self.bins.items():
                yield key.encode('utf-8')
                yield np.require(val, requirements='C').data.cast('b')
        yield self.__class__.__name__.encode('utf-8')
        yield self.name.encode('utf-8')
        yield self.what.encode('utf-8')

    def _check_datasets_consistency(self, other, operation=""):
        if other.shape != self.shape:
            raise ValueError(f"Datasets to {operation} do not have same shape")
        if other.bins != OrderedDict():
            if any((s != o) for s, o in zip(self.bins, other.bins)):
                raise ValueError(f"Datasets to {operation} do not have same "
                                 "bin names")
            if not all(np.array_equal(self.bins[s], other.bins[o])
                       for s, o in zip(self.bins, other.bins)):
                raise ValueError(f"Datasets to {operation} do not have the "
                                 "same bins")

    def __add__(self, other):
        LOGGER.debug("in %s.__add__", self.__class__.__name__)
        if not isinstance(other, (int, float, np.ndarray, Dataset)):
            raise TypeError("Int, float, np.array and Dataset"
                            "accepted for the moment")
        if not isinstance(other, Dataset):
            return Dataset(self.value + other, self.error,
                           bins=self.bins, name=self.name, what=self.what)
        self._check_datasets_consistency(other, "add")
        value = self.value + other.value
        error = np.sqrt(self.error**2 + other.error**2)
        what = (self.what if other.what == self.what
                else self.what+'+'+other.what)
        return Dataset(value, error, bins=self.bins, name=self.name, what=what)

    def __sub__(self, other):
        LOGGER.debug("in %s.__sub__", self.__class__.__name__)
        if not isinstance(other, (int, float, np.ndarray, Dataset)):
            raise TypeError("Int, float, np.array and Dataset "
                            "accepted for the moment")
        if not isinstance(other, Dataset):
            return Dataset(self.value - other, self.error,
                           bins=self.bins, name=self.name, what=self.what)
        self._check_datasets_consistency(other, "subtract")
        value = self.value - other.value
        error = np.sqrt(self.error**2 + other.error**2)
        what = (self.what if other.what == self.what
                else self.what+'-'+other.what)
        return Dataset(value, error, bins=self.bins, name=self.name, what=what)

    def __mul__(self, other):
        LOGGER.debug("in %s.__mul__", self.__class__.__name__)
        if not isinstance(other, Dataset):
            return Dataset(
                self.value * other, self.error * other,
                bins=self.bins, name=self.name, what=self.what)
        self._check_datasets_consistency(other, "multiply")
        value = self.value * other.value
        error = np.sqrt((self.error * other.value)**2
                        + (other.error * self.value)**2)
        return Dataset(value, error, bins=self.bins, name=self.name,
                       what=self.what+'*'+other.what)

    def __truediv__(self, other):
        LOGGER.debug("in %s.__truediv__", self.__class__.__name__)
        if not isinstance(other, Dataset):
            return Dataset(
                self.value / other, self.error / other,
                bins=self.bins, name=self.name, what=self.what)
        self._check_datasets_consistency(other, "divide")
        value = self.value / other.value
        # RuntimeWarning can be ignored thanks to the commented line.
        # 'log' can be used instead of 'ignore' but did not work.
        # with np.errstate(divide='divide', invalid='ignore'):
        error = np.sqrt((self.error / other.value)**2
                        + (self.value * other.error / other.value**2)**2)
        return Dataset(value, error, bins=self.bins, name=self.name,
                       what=self.what+'/'+other.what)

    @staticmethod
    def _get_bins_slice(index):
        if index.stop is not None:
            stop = index.stop+1 if index.stop > 0 else index.stop
            tmpind = slice(index.start, stop, index.step)
            return tmpind
        return index

    def _get_bins_items(self, index):
        nbins = self.bins.copy()
        slices = index if isinstance(index, tuple) else (index,)
        for ind, kbin, dim in zip(slices, self.bins, self.shape):
            bindex = (ind if len(self.bins[kbin]) == dim
                      else self._get_bins_slice(ind))
            nbins[kbin] = self.bins[kbin][bindex]
        return nbins

    def __getitem__(self, index):
        LOGGER.debug("in %s.__getitem__ with index=%s of type %s",
                     self.__class__.__name__, index, type(index))
        if not isinstance(self.value, np.ndarray):
            raise TypeError("[] (__getitem__) can only be applied "
                            "on numpy.ndarrays")
        if not (isinstance(index, slice)
                or (isinstance(index, tuple)
                    and all(isinstance(i, slice) for i in index))):
            raise TypeError("Index can only be a slice or a tuple of slices")
        if ((isinstance(index, tuple) and self.value.ndim != len(index))
                or (isinstance(index, slice) and self.value.ndim != 1)):
            raise ValueError("len(index) should have the same dimension as "
                             "the value numpy.ndarray, i.e. (# ',' = dim-1).\n"
                             "':' can be used for a slice (dimension) not "
                             "affected by the selection.\n"
                             "Slicing is only possible if ndim == 1")
        value = self.value[index]
        error = self.error[index]
        bins = self._get_bins_items(index)
        LOGGER.debug("Shape: %s -> %s", self.shape, value.shape)
        LOGGER.debug("Bins:%s -> %s", self.bins, bins)
        return Dataset(value, error, bins=bins, name=self.name, what=self.what)

    def mask(self, mask):
        '''Apply a mask to the Dataset value and error.

        Value and error become in that case masked_arrays instead of usual
        arrays, calcuations are preserved, not using the masked elements.

        :rtype: Dataset
        '''
        return Dataset(value=np.ma.masked_array(self.value, mask),
                       error=np.ma.masked_array(self.error, mask),
                       bins=self.bins.copy(),
                       name=self.name, what=self.what)


def consistent_datasets(dss1, dss2):
    '''Return `True` if datasets are consistent = same shape.'''
    return dss1.shape == dss2.shape


def same_coords(ds1, ds2):
    '''Return `True` if coordinates (bins) are compatible.

    :param ds1: the first array of coordinate arrays.
    :param ds2: the second array of coordinate arrays.

    Comparison on keys and values.
    '''
    if not consistent_datasets(ds1, ds2):
        return False
    if len(ds1.bins) != len(ds2.bins):
        return False
    return all((s == o and np.array_equal(ds1.bins[s], ds2.bins[o]))
               for s, o in zip(ds1.bins, ds2.bins))



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

'''Verbosity module.

4 levels of verbosity exist to build tables and plots: SILENT, SUMMARY,
INTERMEDIATE and FULL_DETAILS.
'''
from enum import Enum


class Verbosity(Enum):
    '''Verbosity enum.

    Six levels are currently available: SILENT, SUMMARY, DEFAULT,
    INTERMEDIATE, FULL_DETAILS, DEVELOPMENT.
    '''
    SILENT = 0
    SUMMARY = 1
    DEFAULT = 2
    INTERMEDIATE = 3
    FULL_DETAILS = 4
    DEVELOPMENT = 5



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

'''Module for the ``run`` subcommand.'''

import logging
from pathlib import Path

from ..common import JobCommand, read_env, write_env, build_graphs
from ...cosette.backends.queue import QueueScheduling
from ...chrono import Chrono
from ...cosette.scheduler import Scheduler
from ...cosette.task import TaskStatus
from ...path import ensure


LOGGER = logging.getLogger(__name__)


class RunCommand(JobCommand):
    '''Command class for the ``run`` subcommand.'''

    NAME = 'run'

    HELP = 'Run the tasks defined in a job file.'

    DESC = ('Run the tasks defined by the job() function in JOB_FILE, passing '
            'the given arguments (JOB_ARG) and keyword arguments (-k option) '
            'to job().')

    def register(self, parser):
        '''Register options for this command in the parser.'''
        super().register(parser)
        parser.add_argument('-j', '--workers', action='store', default=4,
                            type=int,
                            help='number of workers to use in parallel')
        parser.add_argument('--env-filename', action='store',
                            default='valjean.env',
                            help='name of the files that contain the '
                            'persistent environment (default: valjean.env)')
        parser.add_argument('--env-format', action='store',
                            choices=('pickle',), default='pickle',
                            help='environment persistency format')
        parser.set_defaults(func=self.execute)

    def execute(self, args, config):
        '''Execute the ``run`` command.'''

        with Chrono() as chrono:
            hard_graph, soft_graph = build_graphs(args)
        LOGGER.info('graphs built in %s seconds', chrono)
        LOGGER.info('hard_graph contains %d tasks', len(hard_graph))
        LOGGER.info('soft_graph contains %d tasks', len(soft_graph))
        LOGGER.info('will schedule up to %d tasks in parallel', args.workers)

        # store the args in the config, for later retrieval
        config['args'] = vars(args)

        output_root = config.query('path', 'output-root')
        # we extract the tasks from the hard-dependency graph; both graphs
        # contain the same nodes anyway
        tasks = hard_graph.nodes()
        task_names = [task.name for task in tasks]
        env = read_env(root=output_root, names=task_names,
                       filename=args.env_filename, fmt=args.env_format)
        new_env = schedule(hard_graph=hard_graph, soft_graph=soft_graph,
                           env=env, config=config, workers=args.workers)

        self.task_diagnostics(tasks=tasks,
                              env=new_env, config=config)

        write_env(env, filename=args.env_filename, fmt=args.env_format)
        return new_env

    @classmethod
    def task_diagnostics(cls, *, tasks, env, config):
        '''Emit diagnostic messages about the status of the tasks. Count how
        many have succeeded, how many have failed, etc.  If any tasks have
        failed, this method writes their names in a file called 'failed_tasks'
        in the log directory.

        :param tasks: the tasks that have been scheduled.
        :type tasks: list(Task)
        :param Env env: the environment.
        :param Config config: the configuration object.
        '''
        from collections import Counter
        count_status = Counter()
        missing = []
        failed = []
        for task in tasks:
            task_name = task.name
            if task_name in env:
                status = env[task.name]['status']
                count_status[status] += 1
                if status == TaskStatus.FAILED:
                    failed.append(task_name)
            else:
                missing.append(task_name)

        if missing:
            LOGGER.warning('the following %s tasks are missing from the '
                           'environment: %s',
                           len(missing), '\n  '.join(missing))

        total_graph = len(tasks)
        msgs = (f'{status.name:>7}: {count}/{total_graph} '
                f'({100*count/total_graph:5.1f}%)'
                for status, count in count_status.items())
        LOGGER.info('final environment statistics:\n  %s',
                    '\n  '.join(msgs))

        cls.write_failed_tasks(failed=failed, config=config)

    @classmethod
    def write_failed_tasks(cls, *, failed, config):
        '''Write the names of the failed tasks in the ``failed_tasks`` file.

        :param list(str) failed: the names of the failed tasks.
        :param Config config: the configuration object.
        '''
        if not failed:
            return
        log_root = config.query('path', 'log-root')
        ensure(log_root, is_dir=True)
        failed_fname = Path(log_root, 'failed-tasks')
        with failed_fname.open('w', encoding='utf-8') as failed_file:
            for task in failed:
                failed_file.write(task + '\n')


def schedule(*, hard_graph, soft_graph, env, config=None, workers=1):
    '''Schedule a graph for execution.

    '''
    scheduler = Scheduler(hard_graph=hard_graph, soft_graph=soft_graph,
                          backend=QueueScheduling(workers))
    new_env = scheduler.schedule(env=env, config=config)
    LOGGER.debug('resulting environment: %s', new_env)
    return new_env
